from ewokscore.task import Task
from ..io.input_stack import input_context
from ..io.output_stack import output_context, OutputStackNumpy
from ..transformation import apply_transformations
from ..evaluation.evaluation import post_evaluation, pre_evaluation


class Reg2DPreEvaluation(
    Task,
    input_names=["imagestacks"],
    optional_input_names=[
        "inputs_are_stacks",
        "stack_names",
        "preferences",
        "relevant_stacks",
    ],
    output_names=["full_ranking", "ranking", "stack_names"],
):
    """Given several different stacks of images requiring the same alignment, create a ranking of which stacks might be the most suitable to get a correct alignment."""

    def run(self):
        preferences = self.get_input_value("preferences", "")
        stack_names = self.get_input_value("stack_names", [])

        # parse the preferences
        preferred_stacks = []
        for elem in preferences.replace(" ", "").split(","):
            if elem == "":
                continue
            try:
                preferred_stacks.append(int(elem))
            except ValueError:
                if elem not in stack_names:
                    raise ValueError(
                        "The preferred stacks must be specified as a comma-seperated list of either the index or the name(if given) of the stack"
                    )
                preferred_stacks.append(stack_names.index(elem))

        # evaluate all stacks and create ranking
        with input_context(
            self.inputs.imagestacks,
            inputs_are_stacks=self.get_input_value("inputs_are_stacks", None),
        ) as stack:
            order = list(pre_evaluation(stack))
            for i in preferred_stacks:
                order.remove(i)
            preferred_stacks.extend(order)

            relevant_stacks = self.get_input_value("relevant_stacks", len(stack))
            self.outputs.ranking = preferred_stacks[:relevant_stacks]
            self.outputs.full_ranking = preferred_stacks
            if stack_names:
                self.outputs.stack_names = [
                    stack_names[i] for i in preferred_stacks[:relevant_stacks]
                ]


class Reg2DPostEvaluation(
    Task,
    input_names=["imagestacks", "transformations"],
    optional_input_names=[
        "url",
        "inputs_are_stacks",
        "chosen_stack",
    ],
    output_names=["imagestacks", "imagestack", "transformations"],
):
    """Given several stacks of images requiring the same alignment and their calculated transformations, determine the stack and list of transformations that resulted in the best registration."""

    def run(self):
        url = self.get_input_value("url", None)
        chosen_stack = self.get_input_value("chosen_stack", -1)
        with output_context(url=url) as ostack:
            with input_context(
                self.inputs.imagestacks,
                inputs_are_stacks=self.get_input_value("inputs_are_stacks", None),
            ) as istacks:
                nstacks = len(self.inputs.transformations)
                if nstacks != istacks.shape[0]:
                    raise ValueError(
                        f"Got {nstacks} lists of transformations and {istacks.shape[0]} imagestacks"
                    )
                transformed = list()
                for index in range(nstacks):
                    out = list()
                    apply_transformations(
                        istacks[index],
                        OutputStackNumpy(out),
                        self.inputs.transformations[index],
                    )
                    transformed.append(out)

                post_eval_rank = post_evaluation(
                    transformed, self.inputs.transformations
                )
                if chosen_stack > -1 and chosen_stack < nstacks:
                    ostack.add_points(transformed[chosen_stack])
                else:
                    ostack.add_points(transformed[post_eval_rank[0]])
            if url:
                self.outputs.imagestack = url
            else:
                self.outputs.imagestack = ostack.data

            self.outputs.imagestacks = transformed
            if chosen_stack < nstacks and chosen_stack > -1:
                self.outputs.transformations = self.inputs.transformations[chosen_stack]
            else:
                self.outputs.transformations = self.inputs.transformations[
                    post_eval_rank[0]
                ]



import numpy
from ...math import fit1d


def test_leastsq():
    nx = 501
    x = numpy.arange(nx)
    x0 = 10
    sx = nx // 4
    A = 1000.0
    p1 = numpy.array([x0, sx, A], dtype=numpy.float32)
    x0, sx, A = tuple(p1)

    data = fit1d.gaussian(x, x0, sx, A)

    p2, _ = fit1d.fitgaussian(x, data)
    numpy.testing.assert_allclose(p1, p2)


def test_linfit():
    x = numpy.asarray(
        [
            1.47,
            1.50,
            1.52,
            1.55,
            1.57,
            1.60,
            1.63,
            1.65,
            1.68,
            1.70,
            1.73,
            1.75,
            1.78,
            1.80,
            1.83,
        ]
    )
    y = numpy.asarray(
        [
            52.21,
            53.12,
            54.48,
            55.84,
            57.20,
            58.57,
            59.93,
            61.29,
            63.11,
            64.47,
            66.28,
            68.10,
            69.92,
            72.19,
            74.46,
        ]
    )
    m = 61.272
    b = -39.062
    vm = 3.1539
    vb = 8.63185
    (m1, b1), (em1, eb1) = fit1d.linfit(x, y, errors=True)
    (m2, b2), (em2, eb2) = fit1d.linfit2(x, y, errors=True)
    vm1 = em1 * em1
    vm2 = em2 * em2
    vb1 = eb1 * eb1
    vb2 = eb2 * eb2
    numpy.testing.assert_array_almost_equal([m, b], [m1, b1], decimal=3)
    numpy.testing.assert_array_almost_equal([m, b], [m2, b2], decimal=3)
    numpy.testing.assert_array_almost_equal([vm, vm], [vm1, vm2], decimal=3)
    numpy.testing.assert_array_almost_equal([vb, vb], [vb1, vb2], decimal=4)
    numpy.testing.assert_allclose([m1, b1], [m2, b2])

    m1, em1 = fit1d.linfit_zerointercept(x, y, errors=True)
    m2, em2 = fit1d.linfit_zerointercept2(x, y, errors=True)
    numpy.testing.assert_allclose(m1, m2)
    numpy.testing.assert_almost_equal(em1, em2, decimal=3)  # em2 maybe not correct???



from ..transformation.homography import (
    matrix_from_params,
    params_from_trans,
    Homography,
)
import numpy


def test_parameter_conversion():
    matrix = numpy.identity(3)
    trafo = Homography(matrix)
    numpy.testing.assert_allclose(
        trafo.passive,
        matrix_from_params(params_from_trans(trafo), trafo._type.identity),
    )

    matrix[0:2, 2] = [1, 2]
    trafo = Homography(matrix)
    numpy.testing.assert_allclose(
        trafo.passive,
        matrix_from_params(params_from_trans(trafo), trafo._type.translation),
    )

    matrix[0, 0] = numpy.cos(0.1)
    matrix[1, 0] = numpy.sin(0.1)
    matrix[0, 1] = -numpy.sin(0.1)
    matrix[1, 1] = numpy.cos(0.1)
    trafo = Homography(matrix)
    numpy.testing.assert_allclose(
        trafo.passive,
        matrix_from_params(params_from_trans(trafo), trafo._type.rigid),
    )

    matrix[1, 1] *= 1.1
    trafo = Homography(matrix)
    numpy.testing.assert_allclose(
        trafo.passive, matrix_from_params(params_from_trans(trafo), trafo._type.affine)
    )

    matrix[2, 0] = 0.1
    trafo = Homography(matrix)
    numpy.testing.assert_allclose(
        trafo.passive,
        matrix_from_params(params_from_trans(trafo), trafo._type.projective),
    )



import numpy
from .base import Features


class SilxDescriptorFeatures(
    Features, registry_id=Features.RegistryId("Descriptor", "Silx")
):
    def __init__(self, silx_features: numpy.recarray) -> None:
        self._silx_features = silx_features
        self._coordinates = numpy.array([self._silx_features.y, self._silx_features.x])
        self._nfeatures = len(self._silx_features)
        super().__init__()

    def __getitem__(self, idx):
        return type(self)(self._silx_features[idx])

    @property
    def silx_features(self) -> numpy.recarray:
        return self._silx_features

    @property
    def coordinates(self) -> numpy.ndarray:
        return self._coordinates

    @property
    def ndim(self) -> int:
        return 2

    @property
    def nfeatures(self) -> int:
        return self._nfeatures



import sys
from numbers import Integral
from typing import Tuple, Sequence, Union, Optional, Iterator
from contextlib import contextmanager
from collections.abc import Sequence as AbsSequence

import numpy
from numpy.typing import DTypeLike

from silx.io import h5py_utils
from silx.io.url import DataUrl


class InputStack(AbsSequence):
    def __enter__(self) -> "InputStack":
        return self

    def __exit__(self, *args) -> None:
        pass

    def __getitem__(self, idx) -> numpy.ndarray:
        raise NotImplementedError

    @property
    def shape(self) -> Tuple[int]:
        raise NotImplementedError

    @property
    def dtype(self) -> DTypeLike:
        raise NotImplementedError

    def __len__(self) -> int:
        return self.shape[0]

    @property
    def ndim(self) -> int:
        return len(self.shape)


class InputStackNumpy(InputStack):
    def __init__(
        self,
        arrays: Union[numpy.ndarray, Sequence[numpy.ndarray]],
        inputs_are_stacks: Optional[bool] = None,
    ) -> None:
        if isinstance(arrays, numpy.ndarray):
            if inputs_are_stacks is None:
                inputs_are_stacks = True
            if not inputs_are_stacks:
                arrays = arrays[None, ...]
            self._array = arrays
        elif isinstance(arrays, Sequence):
            if inputs_are_stacks is None:
                inputs_are_stacks = False
            if inputs_are_stacks:
                self._array = numpy.vstack(arrays)
            else:
                self._array = numpy.stack(arrays, axis=0)
        else:
            raise TypeError("arrays")
        super().__init__()

    def __getitem__(self, idx) -> numpy.ndarray:
        return self._array[idx]

    @property
    def shape(self) -> Tuple[int]:
        return self._array.shape

    @property
    def dtype(self) -> DTypeLike:
        return self._array.dtype


class InputStackHdf5(InputStack):
    def __init__(
        self, uris: Union[str, Sequence[str]], inputs_are_stacks: Optional[bool] = None
    ) -> None:
        if isinstance(uris, str):
            self._uris = [DataUrl(uris)]
            if inputs_are_stacks is None:
                inputs_are_stacks = True
        else:
            self._uris = [DataUrl(s) for s in uris]
            if inputs_are_stacks is None:
                inputs_are_stacks = False
        self._uris_are_stacks = inputs_are_stacks
        self._clean_resources()
        super().__init__()

    def __enter__(self) -> "InputStackHdf5":
        self._clean_resources()
        try:
            for uri in self._uris:
                ctx = h5py_utils.File(uri.file_path())
                self._file_objs.append(ctx.__enter__())
        except BaseException:
            for ctx in self._file_objs:
                ctx.__exit__(*sys.exc_info())
            raise
        return super().__enter__()

    def __exit__(self, *args) -> None:
        for ctx in self._file_objs:
            ctx.__exit__(*args)
        self._clean_resources()

    def _clean_resources(self):
        self._file_objs = list()
        self._dset_objs = list()
        self._cumlen_uris = list()
        self._shape = None
        self._dtype = None

    def __getitem__(self, idx) -> numpy.ndarray:
        if not self._dset_objs:
            self._get_dset_info()

        if isinstance(idx, Tuple):
            idx0 = idx[0]
            idxrest = idx[1:]
        else:
            idx0 = idx
            idxrest = tuple()

        dim0scalar = False
        n = len(self)
        if isinstance(idx0, Integral):
            idx0 = numpy.asarray([idx0])
            dim0scalar = True
        elif isinstance(idx0, slice):
            idx0 = numpy.array(range(*idx0.indices(n)))
        elif isinstance(idx0, Sequence):
            idx0 = numpy.asarray(idx0)
        elif idx0 is Ellipsis:
            idx0 = numpy.array(range(n))
        else:
            raise TypeError

        result = list()
        for i in idx0:
            while i < 0:
                i += n
            dseti = numpy.where(self._cumlen_uris > i)[0][0]
            dset = self._dset_objs[dseti]
            if self._uris_are_stacks:
                if dseti > 0:
                    i -= self._cumlen_uris[dseti - 1]
                idxdset = (i,) + idxrest
            else:
                idxdset = idxrest
            result.append(dset[idxdset])

        if dim0scalar:
            return result[0]
        else:
            return numpy.array(result)

    @property
    def shape(self) -> Tuple[int]:
        if self._shape is None:
            self._get_dset_info()
        return self._shape

    @property
    def dtype(self) -> DTypeLike:
        if self._dtype is None:
            self._get_dset_info()
        return self._dtype

    def _get_dset_info(self):
        if len(self._file_objs) != len(self._uris):
            raise RuntimeError("enter the context first")
        len_uris = list()
        dset_objs = list()
        ndim = set()
        shape = set()
        dtype = set()
        for uri, file_obj in zip(self._uris, self._file_objs):
            dset = file_obj[uri.data_path()]
            if self._uris_are_stacks:
                dset_ndim = dset.ndim
                dset_shape = dset.shape[1:]
                dset_len = len(dset)
            else:
                dset_ndim = dset.ndim + 1
                dset_shape = dset.shape
                dset_len = 1

            ndim.add(dset_ndim)
            if len(ndim) != 1:
                raise ValueError("array dimensions do not match")
            shape.add(dset_shape)
            if len(shape) != 1:
                raise ValueError("array dimensions do not match")
            dtype.add(dset.dtype)
            if len(dtype) != 1:
                raise ValueError("array dtype must be the same")
            len_uris.append(dset_len)
            dset_objs.append(dset)
        self._cumlen_uris = numpy.cumsum(len_uris, dtype=int)
        self._shape = (self._cumlen_uris[-1],) + next(iter(shape))
        self._dtype = next(iter(dtype))
        self._dset_objs = dset_objs


@contextmanager
def input_context(
    images, inputs_are_stacks: Optional[bool] = None
) -> Iterator[InputStack]:
    if isinstance(images, str) or (
        isinstance(images, Sequence) and isinstance(images[0], str)
    ):
        with InputStackHdf5(images, inputs_are_stacks=inputs_are_stacks) as stack:
            yield stack
    else:
        with InputStackNumpy(images, inputs_are_stacks=inputs_are_stacks) as stack:
            yield stack



import numpy
from ..transformation.base import Transformation
from ..transformation.types import TransformationType
from ..registry import Registered


__all__ = ["IntensityMapping"]


class IntensityMapping(Registered, register=False):
    SUPPORTED_TRANSFORMATIONS = []

    def __init__(self, transfo_type: TransformationType, **_) -> None:
        if transfo_type in self.SUPPORTED_TRANSFORMATIONS:
            self._type = TransformationType(transfo_type)
        else:
            raise ValueError(
                f"{transfo_type} is not supported by {self.__class__.__name__}"
            )
        super().__init__()

    @property
    def type(self) -> TransformationType:
        return self._type

    def identity(self, dimension: int = 2) -> Transformation:
        raise NotImplementedError

    def calculate(
        self, from_image: numpy.ndarray, to_image: numpy.ndarray
    ) -> Transformation:
        raise NotImplementedError



import numpy as np
from typing import List

from ..io.input_stack import InputStack
from .eval_metrics import peak_eval, smoothness_eval, mse_eval
from ..transformation.homography import Homography


def pre_evaluation(stacks: InputStack):
    """
    Evaluates what the most successful stacks for registration might be.

    param stacks: list of imagestacks of the same image

    returns: indices of the most promising stacks for registration
    """
    evals = list()
    for stack in stacks:
        evals.append(peak_eval(stack, 0))
    return np.argsort(evals)


def post_evaluation(
    aligned_stacks: List[np.ndarray],
    transformations: List[List[Homography]],
):
    """
    Evaluation of the stack after alignment based on the mse error and the smoothness of the transformations

    param aligned_stacks: list of stacks after alignment
    param transformations: list of list of calculated transformations
    param a: the ratio between the two evaluations: total_eval = mse_eval + smoothness_eval

    returns: ranking of presumably best alignments in descending order
    """
    errors = list()
    for aligned_stack, transformation in zip(aligned_stacks, transformations):
        err1 = mse_eval(aligned_stack, 0)
        err2 = smoothness_eval(transformation, aligned_stack[0].shape)
        errors.append(err1 + err2)

    return np.argsort(errors)



from typing import Optional, Iterator, List
from contextlib import contextmanager
from numpy.typing import ArrayLike

from silx.io import h5py_utils
from silx.io.url import DataUrl
from ewoksdata.data.hdf5.dataset_writer import DatasetWriter


class OutputStack:
    def __enter__(self) -> "OutputStack":
        return self

    def __exit__(self, *args) -> None:
        pass

    def add_point(self, data: ArrayLike) -> bool:
        raise NotImplementedError

    def add_points(self, data: ArrayLike) -> bool:
        raise NotImplementedError


class OutputStackNumpy(OutputStack):
    def __init__(self, data: list) -> None:
        self._data = data
        super().__init__()

    @property
    def data(self) -> List[ArrayLike]:
        return self._data

    def add_point(self, data: ArrayLike) -> bool:
        self._data.append(data)

    def add_points(self, data: ArrayLike) -> bool:
        self._data.extend(data)


class OutputStackHdf5(OutputStack):
    def __init__(self, uri: str) -> None:
        self._file_obj = None
        self._writer = None
        self._uri = DataUrl(uri)
        super().__init__()

    def __enter__(self) -> "OutputStackHdf5":
        ctx = h5py_utils.File(self._uri.file_path(), mode="a")
        self._file_obj = ctx.__enter__()

        data_path = self._uri.data_path()
        parent = self._file_obj
        names = [s for s in data_path.split("/") if s]
        if not names:
            raise ValueError("URL needs a data path")
        for name in names[:-1]:
            if not name:
                continue
            parent = parent.require_group(name)

        self._writer = DatasetWriter(parent, names[-1])
        return super().__enter__()

    def __exit__(self, *args) -> None:
        self._writer.flush_buffer()
        self._file_obj.__exit__(*args)
        self._file_obj = None
        self._writer = None
        return super().__exit__(*args)

    def add_point(self, data: ArrayLike) -> bool:
        if self._writer is None:
            raise RuntimeError("enter the context first")
        self._writer.add_point(data)

    def add_points(self, data: ArrayLike) -> bool:
        if self._writer is None:
            raise RuntimeError("enter the context first")
        self._writer.add_points(data)


@contextmanager
def output_context(url: Optional[str] = None) -> Iterator[OutputStack]:
    if url:
        with OutputStackHdf5(url) as stack:
            yield stack
    else:
        data = list()
        with OutputStackNumpy(data) as stack:
            yield stack



import numpy
from .lstsq import get_lstsq_solver
from .base import FeatureMapping
from ..detection.base import Features
from ...transformation import TransformationType
from ...transformation.numpy_backend import NumpyHomography

__all__ = ["NumpyLstSqFeatureMapping"]


class NumpyLstSqFeatureMapping(
    FeatureMapping, registry_id=FeatureMapping.RegistryId("LstSq", "Numpy")
):
    def __init__(self, transfo_type: TransformationType) -> None:
        self._solver = get_lstsq_solver(transfo_type)
        self._transfo_type = transfo_type
        super().__init__()

    def calculate(
        self, from_features: Features, to_features: Features
    ) -> NumpyHomography:
        passive = self._solver(to_features.coordinates, from_features.coordinates)
        return NumpyHomography(passive, self._transfo_type)

    def identity(self) -> NumpyHomography:
        return NumpyHomography(numpy.identity(3), TransformationType.identity)



import numpy
import pytest
from ..features import registration
from ..io.input_stack import InputStackNumpy
from .data import data_for_registration


_DETECTORS = {
    f"detector{'_'.join(k)}": v
    for k, v in registration.FeatureDetector.get_subclass_items()
}
_MATCHERS = {
    f"matcher{'_'.join(k)}": v
    for k, v in registration.FeatureMatching.get_subclass_items()
}
_MAPPERS = {
    f"mapper{'_'.join(k)}": v
    for k, v in registration.FeatureMapping.get_subclass_items()
}


@pytest.mark.parametrize("mapper", _MAPPERS)
@pytest.mark.parametrize("detector", _DETECTORS)
@pytest.mark.parametrize("matcher", _MATCHERS)
@pytest.mark.parametrize("transfo_type", ["translation"])
def test_feature_registration(transfo_type, matcher, detector, mapper):
    istack, active1, passive1 = data_for_registration.images(transfo_type, plot=0)
    istack = InputStackNumpy(istack)

    detector = _DETECTORS[detector]()
    matcher = _MATCHERS[matcher]()
    mapper = _MAPPERS[mapper](transfo_type)

    transformations = registration.calculate_transformations(
        istack, detector, matcher, mapper
    )

    active2 = numpy.stack([tr.active for tr in transformations])
    passive2 = numpy.stack([tr.passive for tr in transformations])

    if "Sift" in detector.get_subclass_id():
        rtol = 0.01
    else:
        rtol = 0.05
    numpy.testing.assert_allclose(active1, active2, rtol=rtol)
    numpy.testing.assert_allclose(passive1, passive2, rtol=rtol)



from contextlib import contextmanager
from AnyQt import QtWidgets

from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.bindings import ow_build_opts
from ewoksorange.gui.parameterform import ParameterForm

from ewoksndreg.tasks import Reg2DTransform
from ewoksndreg.io.input_stack import input_context, InputStack

__all__ = ["OWReg2DTransform"]


class OWReg2DTransform(
    OWEwoksWidgetOneThread, **ow_build_opts, ewokstaskclass=Reg2DTransform
):
    name = "2D Transformation"
    description = "Apply image transformations to a stack of images"
    icon = "icons/transformation.png"

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._init_control_area()
        self._init_main_area()

    def _init_control_area(self):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        values = self.get_default_input_values(
            include_missing=True, defaults={"interpolation_order": 1}
        )

        options = {
            "imagestack": {"value_for_type": "", "select": "h5dataset"},
            "inputs_are_stacks": {"value_for_type": False},
            "url": {"value_for_type": ""},
            "interpolation_order": {"value_for_type": 0},
            "crop": {"value_for_type": False},
        }

        for name, kw in options.items():
            if name not in options:
                continue
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_input_data()

    def handleNewSignals(self) -> None:
        self._update_input_data()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_output_data()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()

        self._tabs = QtWidgets.QTabWidget(parent=self.mainArea)
        layout.addWidget(self._tabs)

        w = QtWidgets.QWidget(parent=self.mainArea)
        layout = QtWidgets.QVBoxLayout()
        w.setLayout(layout)
        self._oplot = Plot2D(parent=w)
        self._oplot.setDefaultColormap(Colormap("viridis"))
        self._oslider1 = HorizontalSliderWithBrowser(parent=w)
        self.current_in_stack = 0
        self._oslider2 = HorizontalSliderWithBrowser(parent=w)
        self.current_in_image = 0
        layout.addWidget(self._oplot)
        layout.addWidget(self._oslider1)
        layout.addWidget(self._oslider2)
        self._tabs.addTab(w, "Aligned")

        w = QtWidgets.QWidget(parent=self.mainArea)
        layout = QtWidgets.QVBoxLayout()
        w.setLayout(layout)
        self._iplot = Plot2D(parent=w)
        self._iplot.setDefaultColormap(Colormap("viridis"))
        self._islider1 = HorizontalSliderWithBrowser(parent=w)
        self.current_out_stack = 0
        self._islider2 = HorizontalSliderWithBrowser(parent=w)
        self.current_out_image = 0
        layout.addWidget(self._iplot)
        layout.addWidget(self._islider1)
        layout.addWidget(self._islider2)
        self._tabs.addTab(w, "Original")

        self._islider1.valueChanged[int].connect(self._select_in_stack)
        self._islider2.valueChanged[int].connect(self._select_in_image)
        self._oslider1.valueChanged[int].connect(self._select_out_stack)
        self._oslider2.valueChanged[int].connect(self._select_out_image)
        self._update_input_data()

    @contextmanager
    def _input_context(self, images=None):
        if images is not None:
            yield images
            return
        try:
            binit = True
            with input_context(self.get_task_input_value("imagestack")) as images:
                binit = False
                yield images
        except TypeError:
            if binit:
                yield list()

    @contextmanager
    def _output_context(self, images=None):
        if images is not None:
            yield images
            return
        try:
            binit = True
            with input_context(self.get_task_output_value("imagestack")) as images:
                binit = False
                yield images
        except TypeError:
            if binit:
                yield list()

    def _update_input_data(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

        with self._input_context() as images:
            if isinstance(images, InputStack):
                if images.ndim == 3:
                    self._islider1.hide()
                    self._islider2.setMaximum(max(len(images) - 1, 0))
                else:
                    self._islider1.show()
                    self._islider1.setMaximum(max(len(images) - 1, 0))
                    self._islider2.setMaximum(max(len(images[0]) - 1, 0))
            self._select_in_stack(self._islider1.value(), images=images)
            self._select_in_image(self._islider2.value(), images=images)

    def _select_in_stack(self, select, images=None):
        self.current_in_stack = select
        self._select_in_image(self.current_in_image, images)

    def _select_in_image(self, select, images=None):
        with self._input_context(images=images) as images:
            if images:
                self.current_in_image = select
                if images.ndim == 3:
                    self._iplot.addImage(images[select], legend="image")
                else:
                    self._iplot.addImage(
                        images[self.current_in_stack, select], legend="image"
                    )

    def _update_output_data(self):
        with self._output_context() as images:
            if isinstance(images, InputStack):
                if images.ndim == 3:
                    self._oslider1.hide()
                    self._oslider2.setMaximum(max(len(images) - 1, 0))
                else:
                    self._oslider1.show()
                    self._oslider1.setMaximum(max(len(images) - 1, 0))
                    self._oslider2.setMaximum(max(len(images[0]) - 1, 0))
            self._select_out_stack(self._oslider1.value(), images=images)
            self._select_out_image(self._oslider2.value(), images=images)

    def _select_output_image(self, select, images=None):
        with self._output_context(images=images) as images:
            if images:
                self._oplot.addImage(images[select], legend="image")

    def _select_out_stack(self, select, images=None):
        self.current_out_stack = select
        self._select_out_image(self.current_out_image, images)

    def _select_out_image(self, select, images=None):
        with self._output_context(images=images) as images:
            if images:
                self.current_out_image = select
                if images.ndim == 3:
                    self._oplot.addImage(images[select], legend="image")
                else:
                    self._oplot.addImage(
                        images[self.current_out_stack, select], legend="image"
                    )



from enum import Enum


class SitkMetricType(str, Enum):
    MeanSquares = "MeanSquares"
    CrossCorrelation = "CrossCorrelation"
    ANTSNeighborhoodCorrelation = "ANTSNeighborhoodCorrelation"
    JointHistogramMutualInformation = "JointHistogramMutualInformation"
    MattesMutualInformation = "MattesMutualInformation"


class SitkOptimizerType(str, Enum):
    Exhaustive = "Exhaustive"
    Powell = "Powell"
    Evolutionary = "Evolutionary"
    GradientDescent = "GradientDescent"
    GradientDescentLineSearch = "GradientDescentLineSearch"
    CGLineSearch = "CGLineSearch"
    LBFGSB = "LBFGSB"


class KorniaMetricType(str, Enum):
    MeanAbsoluteError = "MeanAbsoluteError"
    MeanSquaredError = "MeanSquaredError"
    MutualInformation = "MutualInformation"


class KorniaOptimizerType(str, Enum):
    Adam = "Adam"
    SGD = "SGD"
    RMSprop = "RMSprop"



import warnings
import numpy

try:
    from scipy.optimize import leastsq as _leastsq
    from scipy.optimize import nnls
    from scipy.optimize import lsq_linear
except ImportError:
    _leastsq = None
    nnls = None
    lsq_linear = None


def gaussian(x, x0, sx, A):
    return (
        A / (numpy.sqrt(2 * numpy.pi) * sx) * numpy.exp(-((x - x0) ** 2) / (2 * sx**2))
    )


def guess_gaussian(x, data):
    x0i = numpy.argmax(data)
    x0 = x[x0i]
    sx = numpy.sqrt(abs((x - x0) ** 2 * data).sum() / data.sum())
    A = data[x0] * numpy.sqrt(2 * numpy.pi) * sx
    return numpy.array([x0, sx, A], dtype=numpy.float32)


def fitgaussian(x, data):
    return leastsq(x, data, guessfunc=guess_gaussian, fitfunc=gaussian)


def leastsq(x, data, guessfunc=None, fitfunc=None):
    if _leastsq is None:
        raise RuntimeError("requires 'scipy'")
    guess = guessfunc(x, data)

    def errorfunc(p, x, data):
        return numpy.ravel(fitfunc(x, *p) - data)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        p, success = _leastsq(errorfunc, guess, args=(x, data))
        success = success > 0 and success < 5

    return p, success


def xyremovenan(x, y):
    b = numpy.logical_and(~numpy.isnan(x), ~numpy.isnan(y))
    return x[b], y[b]


def cor_from_cov(cov):
    D = numpy.diag(1 / numpy.sqrt(numpy.diag(cov)))
    return D.dot(cov.dot(D))


def lstsq_cov(A, vare=None, cove=None):
    """Covariance matrix of the least squares solution
    of a linear system.

    .. math::

        A.x = b + e

        E(e) = 0

    Args:
        A(array): (m x n)
        vare(array): variance on e (m)
        cove(array): covariance matrix of e (m x m)

    Returns:
        covx(array): (n x n)
    """
    # A.x = b + e
    # E(e) = 0
    #
    # Least squares estimator of x:
    #  x = M.b
    #  M = (A^T.A)^(-1).A^T
    #  COV(x) = M.COV(e).M^T
    # https://stat.ethz.ch/~geer/bsa199_o.pdf

    m, n = A.shape
    if m == n:
        try:
            # M = A^(-1)
            if cove is None:
                return numpy.linalg.inv(A.T.dot(A / vare.reshape((m, 1))))
            else:
                iA = numpy.linalg.inv(A)
                return iA.dot(cove.dot(iA.T))
        except numpy.linalg.linalg.LinAlgError:
            pass
    # TODO: this matrix has been already calculated before
    M = numpy.linalg.inv((A.T.dot(A))).dot(A.T)
    if cove is None:
        return (M * vare.reshape((1, m))).dot(M.T)
    else:
        return M.dot(cove).dot(M.T)


def lstsq_std(A, b=None, x=None, vare=None, cove=None):
    """Estimated error of solution to linear system

    .. math::

        A.x = b + e

        E(e) = 0

    Args:
        A(array): (m x n)
        b(array): only needed when neither vare nor cove are given (m)
        x(array): only needed when neither vare nor cove are given (n)
        vare(array): variance on e (m)
        cove(array): covariance matrix of e (m x m)

    Returns:
        stdx(array): errors (n)
    """
    if vare is None and cove is None:
        vare = numpy.var(numpy.dot(A, x) - b, ddof=x.size)
        vare = numpy.array([vare] * b.size)
    return numpy.sqrt(numpy.diag(lstsq_cov(A, vare=vare, cove=cove)))


def lstsq_std_indep(A, b=None, x=None, vare=None):
    """Estimated error of solution to linear system

    .. math::

        A.x = b + e

        E(e) = 0

    Assume we know x are independent random variables then
    there variances are found by solving another linear system

    .. math::
        (A*A).VAR(x) = VAR(e)

    Args:
        A(array): (m x n)
        b(array): (m)
        x(array): (n)
        vare(array): variance on e (m)

    Returns:
        stdx(array): sqrt(VARX) (n)
    """
    if vare is None:
        vare = numpy.var(numpy.dot(A, x) - b, ddof=x.size)
        vare = numpy.full(vare, x.size)
    return numpy.sqrt(lstsq(A * A, vare))


def lstsq(A, b, errors=False, vare=None, cove=None):
    """Solve the following linear system

    .. math::

        A.x = b + e

        E(e) = 0

    Args:
        A(array): (m x n)
        b(array): (m)
        errors(Optional(bool)): return solution with estimated error
        vare(array): variance on e (m)
        cove(array): covariance matrix of e (m x m)

    Returns:
        x(array): solution (n)
        stdx(array): optional errors (n)
    """
    x = numpy.linalg.lstsq(A, b, rcond=-1)[0]
    if errors:
        return x, lstsq_std(A, b=b, x=x, vare=vare, cove=cove)
    else:
        return x


def lstsq_nonnegative(A, b, errors=False, vare=None):
    """Solve the following linear system

    .. math::

        A.x = b + e \\quad x>=0

        E(e) = 0

    Args:
        A(array): (m x n)
        b(array): (m)
        errors(Optional(bool)): return solution with estimated error
        vare(array): variance on e (m)

    Returns:
        x(array): solution (n)
        stdx(array): optional errors (n)
    """
    if nnls is None:
        raise RuntimeError("requires 'scipy'")
    x = nnls(A, b)[0]
    if errors:
        return x, lstsq_std(A, b=b, x=x, vare=vare)
    else:
        return x


def lstsq_bound(A, b, lb, ub, errors=False, vare=None):
    r"""Solve the following linear system

    .. math::

        A.x = b + e \quad lb<=x<=ub

        E(e) = 0

    Args:
        A(array): (m x n)
        b(array): (m)
        lb(num): lower bound
        ub(num): upper bound
        errors(Optional(bool)): return solution with estimated error
        vare(array): variance on e (m)

    Returns:
        x(array): solution (n)
        stdx(array): optional errors (n)
    """
    if lsq_linear is None:
        raise RuntimeError("requires 'lsq_linear'")
    x = lsq_linear(A, b, bounds=(lb, ub)).x
    if errors:
        return x, lstsq_std(A, b=b, x=x, vare=vare)
    else:
        return x


def linfit(x, y, errors=False, vare=None):
    """Linear fit

    .. math::

        y = m.x + b + e

        E(e) = 0

    Args:
        x(array): (m)
        y(array): (m)
        errors(Optional(bool)): return solution with estimated error
        vare(array): variance on e (m)

    Returns:
        sol(array): solution (m,b)
        stdsol(array): optional errors (2-tuple)
    """
    A = numpy.vstack([x, numpy.ones(len(x))]).T
    return lstsq(A, y, errors=errors, vare=vare)  # slope,intercept


def linfit2(x, y, errors=False, vare=None):
    if vare is not None:
        raise NotImplementedError("Use linfit instead")
    n = len(x)
    Sxy = (x * y).sum()
    Sxx = (x * x).sum()
    Sx = x.sum()
    Sy = y.sum()
    denom = float(n * Sxx - Sx * Sx)
    mnum = n * Sxy - Sx * Sy
    bnum = Sxx * Sy - Sx * Sxy
    m = mnum / denom
    b = bnum / denom

    if errors:
        Syy = (y * y).sum()
        num = n * Syy - Sy * Sy - m * mnum
        mstd = numpy.sqrt(num / ((n - 2.0) * denom))
        bstd = numpy.sqrt(num * Sxx / (n * (n - 2.0) * denom))
        return [m, b], [mstd, bstd]
    else:
        return [m, b]


def nanlinfit(x, y, errors=False, vare=None):
    x, y = xyremovenan(x, y)
    return linfit(x, y, errors=errors, vare=vare)


def nanlinfit2(x, y, errors=False, vare=None):
    x, y = xyremovenan(x, y)
    return linfit2(x, y, errors=errors, vare=vare)


def linfit_zerointercept(x, y, errors=False, vare=None):
    """Linear fit with zero intercept

    .. math::

        y = m.x + e

        E(e) = 0

    Args:
        x(array): (m)
        y(array): (m)
        errors(Optional(bool)): return solution with estimated error
        vare(array): variance on e (m)

    Returns:
        m(array): solution
        stdm(array): optional error
    """
    A = numpy.vstack([x]).T
    if errors:
        m, mstd = lstsq(A, y, errors=True, vare=vare)
        return m[0], mstd[0]
    else:
        return lstsq(A, y)[0]


def linfit_zerointercept2(x, y, errors=False, vare=None):
    Sxy = (x * y).sum()
    Sxx = float((x * x).sum())
    m = Sxy / Sxx
    if errors:
        n = len(x)
        Syy = (y * y).sum()
        mstd = numpy.sqrt(
            (Syy + m * m * Sxx - 2 * m * Sxy) / ((n - 1.0) * Sxx)
        )  # Not sure
        return m, mstd
    return m



import numpy as np

from ..math.crop import crop_edges, crop_stack


def test_crop_stack():
    array = np.zeros((15, 15, 15))
    array[:, 0:2, :] = np.nan
    cropped = crop_stack(array)
    assert cropped.shape == (15, 13, 15)
    array[10, :, 0] = np.nan
    cropped = crop_stack(array)
    assert cropped.shape == (15, 13, 14)


def test_crop_edges():
    array = np.zeros((15, 15, 15))
    array[:, 0:2, :] = np.nan
    x, y = crop_edges(array)
    assert x == slice(2, 15)
    assert y == slice(0, 15)
    array[10, :, 0] = np.nan
    x, y = crop_edges(array)
    assert x == slice(2, 15)
    assert y == slice(1, 15)



"""Transformations between n-D datasets"""

from .types import TransformationType  # noqa F401
from .base import *  # noqa F401
from .numpy_backend import *  # noqa F401

try:
    from .scipy_backend import *  # noqa F401
except ImportError:
    pass

try:
    from .scikitimage_backend import *  # noqa F401
except ImportError:
    pass

try:
    from .simpleitk_backend import *  # noqa F401
except ImportError:
    pass

from .apply import apply_transformations  # noqa F401



import numpy
import pytest
from ..intensities import registration
from ..io.input_stack import InputStackNumpy
from .data import data_for_registration
from ..transformation.homography import Homography
from ..transformation.numpy_backend import NumpyHomography

_MAPPERS = {
    f"mapper{'_'.join(k)}": v
    for k, v in registration.IntensityMapping.get_subclass_items()
}


@pytest.mark.parametrize("mapper", _MAPPERS)
@pytest.mark.parametrize(
    "transfo_type",
    ["translation", "rigid", "affine", "similarity"],
)
def test_intensity_registration(transfo_type, mapper):
    if transfo_type not in _MAPPERS[mapper].SUPPORTED_TRANSFORMATIONS:
        pytest.skip(f"transformation type {transfo_type} not supported by {mapper}")

    istack, active1, passive1 = data_for_registration.images(transfo_type, plot=0)
    istack = InputStackNumpy(istack)

    mapper = _MAPPERS[mapper](transfo_type=transfo_type)

    transformations = registration.calculate_transformations(istack, mapper)

    if isinstance(transformations[0], Homography):
        active2 = numpy.stack([tr.active for tr in transformations])
        passive2 = numpy.stack([tr.passive for tr in transformations])
        numpy.testing.assert_allclose(active1, active2, rtol=0.1, atol=0.1)
        numpy.testing.assert_allclose(passive1, passive2, rtol=0.1, atol=0.1)

    nx, ny = numpy.mgrid[0:100:10, 0:100:10]
    nx, ny = nx.astype(numpy.float64), ny.astype(numpy.float64)

    for i, transformation in enumerate(transformations):
        res1 = transformation.apply_coordinates(
            numpy.asarray([nx.flatten(), ny.flatten()])
        )
        res2 = NumpyHomography(passive1[i]).apply_coordinates(
            numpy.asarray([nx.flatten(), ny.flatten()])
        )
        numpy.testing.assert_allclose(res1, res2, rtol=0.1, atol=0.4)


@pytest.mark.parametrize("mapper", _MAPPERS)
@pytest.mark.parametrize("transfo_type", ["translation", "affine"])
def test_intensity_registration_block(transfo_type, mapper):
    if transfo_type in _MAPPERS[mapper].SUPPORTED_TRANSFORMATIONS:
        istack, active1, passive1 = data_for_registration.images(
            transfo_type, nimages=7, plot=0
        )
        istack = InputStackNumpy(istack)

        mapper = _MAPPERS[mapper](transfo_type=transfo_type)

        transformations = registration.calculate_transformations(
            istack, mapper, include=[0, 1, 2, 4, 5, 6], reference=0, block_size=3
        )
        active1.pop(3)
        passive1.pop(3)
        if isinstance(transformations[0], Homography):
            active2 = numpy.stack([tr.active for tr in transformations])
            passive2 = numpy.stack([tr.passive for tr in transformations])
            numpy.testing.assert_allclose(active1, active2, rtol=0.01, atol=0.1)
            numpy.testing.assert_allclose(passive1, passive2, rtol=0.01, atol=0.1)

        nx, ny = numpy.mgrid[0:100:10, 0:100:10]
        nx, ny = nx.astype(numpy.float64), ny.astype(numpy.float64)

        for i, transformation in enumerate(transformations):
            res1 = transformation.apply_coordinates(
                numpy.asarray([nx.flatten(), ny.flatten()])
            )
            res2 = NumpyHomography(passive1[i]).apply_coordinates(
                numpy.asarray([nx.flatten(), ny.flatten()])
            )
            numpy.testing.assert_allclose(res1, res2, rtol=0.1, atol=0.4)



from typing import Tuple, Optional, Mapping
from skimage.feature import match_descriptors
from .base import FeatureMatching
from ..features import Features
from ..features import SciKitImageDescriptorFeatures

__all__ = ["SciKitImageDescriptorFeatureMatching"]


class SciKitImageDescriptorFeatureMatching(
    FeatureMatching, registry_id=FeatureMatching.RegistryId("Descriptor", "SciKitImage")
):
    def __init__(self, match_options: Optional[Mapping] = None, **kw) -> None:
        if match_options is None:
            match_options = dict()
        self._match_options = match_options
        super().__init__(**kw)

    def match(
        self, from_features: Features, to_features: Features, *_
    ) -> Tuple[Features, Features]:
        sk_from_features = from_features.as_type(SciKitImageDescriptorFeatures)
        sk_to_features = to_features.as_type(SciKitImageDescriptorFeatures)
        idx_from, idx_to = match_descriptors(
            sk_from_features.skimage_descriptors,
            sk_to_features.skimage_descriptors,
            **self._match_options
        ).T
        return (from_features[idx_from], to_features[idx_to])



from typing import List, Dict
from ewoksorange.bindings import ows_to_ewoks
from ewokscore import execute_graph

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files


def test_reg2d_intensities_without_qt(tmpdir):
    from orangecontrib.ewoksndreg import tutorials

    filename = resource_files(tutorials) / "reg2d_intensities.ows"
    assert_reg2d_intensities_without_qt(filename, tmpdir)


def test_reg2d_intensities_with_qt(ewoks_orange_canvas, tmpdir):
    from orangecontrib.ewoksndreg import tutorials

    filename = resource_files(tutorials) / "reg2d_intensities.ows"
    assert_reg2d_intensities_with_qt(ewoks_orange_canvas, filename, tmpdir)


def assert_reg2d_intensities_without_qt(filename, tmpdir):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    outputs = execute_graph(
        graph, inputs=get_inputs(tmpdir), outputs=[{"all": True}], merge_outputs=False
    )
    expected = get_expected_outputs(tmpdir)
    label_to_id = {
        attrs["label"]: node_id for node_id, attrs in graph.graph.nodes.items()
    }
    outputs = {k: set(v) for k, v in outputs.items()}
    expected = {label_to_id[k]: v for k, v in expected.items()}
    assert outputs == expected


def assert_reg2d_intensities_with_qt(ewoks_orange_canvas, filename, tmpdir):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_graph(str(filename), inputs=get_inputs(tmpdir))
    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=10)
    outputs = dict(ewoks_orange_canvas.iter_output_values())
    outputs = {k: set(v) for k, v in outputs.items()}
    assert outputs == get_expected_outputs(tmpdir)


def get_inputs(tmpdir) -> List[dict]:
    return list()


def get_expected_outputs(tmpdir) -> Dict[str, dict]:
    return {
        "Align (theory)": {"imagestack"},
        "2D Intensity-Based Registration": {"transformations"},
        "Align": {"imagestack"},
        "2D Example Stack": {"imagestack", "transformations"},
    }



from typing import Type, Sequence, Optional
import numpy
from .types import TransformationType
from ..registry import Registered

__all__ = ["Transformation"]


class Transformation(Registered, register=False):
    def __init__(self, transfo_type: TransformationType) -> None:
        if isinstance(transfo_type, str):
            transfo_type = TransformationType(transfo_type)
        self._type = transfo_type

    @property
    def type(self) -> TransformationType:
        return self._type

    def as_type(self, cls: Type["Transformation"]) -> "Transformation":
        if isinstance(self, cls):
            return self
        raise TypeError(f"cannot convert '{type(self).__name__}' to '{cls.__name__}'")

    def is_homography(self):
        return self._type in [
            "identity",
            "translation",
            "rigid",
            "similarity",
            "affine",
            "projective",
        ]

    def apply_coordinates(self, coord: Sequence[numpy.ndarray]) -> numpy.ndarray:
        """
        :param coord: shape `(N, M)`
        :returns: shape `(N, M)`
        """
        raise NotImplementedError

    def apply_data(
        self,
        data: numpy.ndarray,
        offset: Optional[numpy.ndarray] = None,
        shape: Optional[numpy.ndarray] = None,
        cval=numpy.nan,
        interpolation_order: int = 1,
    ) -> numpy.ndarray:
        """
        :param data: shape `(N1, N2, ..., M1, M2, ...)` with `len((N1, N2, ...)) = N`
        :param offset: shape `(N,)`
        :param shape: shape `(N,) = [N1', N2', ...]`
        :param cval: missing value
        :param interpolation_order: order of interpolation: 0 is nearest neighbor, 1 is bilinear,...
        :returns: shape `(N1', N2', ..., M1, M2, ...)`
        """
        raise NotImplementedError

    def __mul__(self, other: "Transformation") -> "Transformation":
        """When appyling the transformation, `other` comes after `self`"""
        raise NotImplementedError



from ewokscore.task import Task
from ..io.input_stack import input_context
from ..intensities import registration

__all__ = ["Reg2DIntensities"]


class Reg2DIntensities(
    Task,
    input_names=["imagestack", "mapper", "transformation_type"],
    optional_input_names=[
        "mask",
        "inputs_are_stacks",
        "block_size",
        "reference",
        "preprocessing_options",
        "method_options",
    ],
    output_names=["transformations"],
):
    """Use an intensity-based registration method to calculate transformations to register the 2D-images in the stack."""

    def run(self):
        preprocessing_options = self.get_input_value("preprocessing_options", {})
        method_options = self.get_input_value("method_options", {})
        mask = self.get_input_value("mask", None)
        mapper = registration.IntensityMapping.get_subclass(self.inputs.mapper)(
            transfo_type=self.inputs.transformation_type, mask=mask, **method_options
        )
        with input_context(
            self.inputs.imagestack,
            inputs_are_stacks=self.get_input_value("inputs_are_stacks", None),
        ) as stack:
            if stack.ndim == 3:
                self.outputs.transformations = registration.calculate_transformations(
                    stack,
                    mapper,
                    reference=self.get_input_value("reference", 0),
                    block_size=self.get_input_value("block_size", 1),
                    preprocessing_options=preprocessing_options,
                )
                return
            if stack.ndim == 4:
                self.outputs.transformations = [
                    registration.calculate_transformations(
                        stack_3d,
                        mapper,
                        reference=self.get_input_value("reference", 0),
                        block_size=self.get_input_value("block_size", 1),
                        preprocessing_options=preprocessing_options,
                    )
                    for stack_3d in stack
                ]
            else:
                raise ValueError("This task only handles 3D or 4D data of 2D images")



"""Match features to determine the mapping between two n-D datasets"""

from .base import *  # noqa F401
from .numpy_backend import *  # noqa F401

try:
    from .silx_backend import *  # noqa F401
except ImportError:
    pass

try:
    from .scikitimage_backend import *  # noqa F401
except ImportError:
    pass



from typing import Optional, Type, Callable
from collections.abc import Sequence as AbcSequence
import numpy
from ...registry import Registered


class Features(AbcSequence, Registered, register=False):
    _FEATURE_TYPE_CONVERTERS = dict()

    @classmethod
    def _set_feature_type_converter(
        cls, to_cls: Type["Features"], converter: Callable[["Features"], "Features"]
    ):
        cls._FEATURE_TYPE_CONVERTERS[(cls, to_cls)] = converter

    @classmethod
    def _get_feature_type_converter(
        cls, to_cls: Type["Features"]
    ) -> Optional[Callable[["Features"], "Features"]]:
        return cls._FEATURE_TYPE_CONVERTERS.get((cls, to_cls))

    def __init_subclass__(cls, **kw) -> None:
        super().__init_subclass__(**kw)
        cls._set_feature_type_converter(cls, lambda x: x)

    def as_type(self, cls: Type["Features"]) -> "Features":
        func = self._get_feature_type_converter(cls)
        if func is not None:
            return func(self)
        raise TypeError(f"cannot convert '{type(self).__name__}' to '{cls.__name__}'")

    @property
    def coordinates(self) -> numpy.ndarray:
        # ndim x nfeatures
        raise NotImplementedError

    @property
    def ndim(self) -> int:
        raise NotImplementedError

    @property
    def nfeatures(self) -> int:
        raise NotImplementedError

    def __len__(self) -> int:
        return self.nfeatures

    def __getitem__(self, idx):
        raise NotImplementedError



from typing import Optional, Sequence, Dict
import numpy
from scipy.ndimage import affine_transform, shift
from .homography import Homography
from .base import Transformation
from .numpy_backend import homography_transform_coordinates

__all__ = ["ScipyHomography"]


class ScipyHomography(
    Homography, registry_id=Homography.RegistryId("Homography", "Scipy")
):
    def __init__(self, *args, warp_options: Optional[Dict] = None, **kw) -> None:
        if warp_options is None:
            warp_options = dict()
        self._warp_options = warp_options
        super().__init__(*args, **kw)

    def apply_coordinates(self, coord: Sequence[numpy.ndarray]) -> numpy.ndarray:
        """
        :param coord: shape `(N, M)`
        :returns: shape `(N, M)`
        """
        return homography_transform_coordinates(self.active, coord)

    def apply_data(
        self,
        data: numpy.ndarray,
        offset: Optional[numpy.ndarray] = None,
        shape: Optional[numpy.ndarray] = None,
        cval=numpy.nan,
        interpolation_order: int = 1,
    ) -> numpy.ndarray:
        """
        :param data: shape `(N1, N2, ..., M1, M2, ...)` with `len((N1, N2, ...)) = N`
        :param offset: shape `(N,)`
        :param shape: shape `(N,) = [N1', N2', ...]`
        :param cval: missing value
        :param interpolation_order: order of interpolation: 0 is nearest neighbor, 1 is bilinear,...
        :returns: shape `(N1', N2', ..., M1, M2, ...)`
        """
        kw = dict(self._warp_options)
        if shape is not None:
            kw["output_shape"] = shape
        if offset is not None:
            kw["offset"] = offset
        if cval is not None:
            kw["cval"] = cval
        if interpolation_order is not None:
            kw["order"] = interpolation_order
        # TODO: offset, shape
        if self.type == self.type.identity:
            return data
        if self.type == self.type.translation:
            return shift(data, -self.passive[:-1, -1], **kw)
        if self.type in (
            self.type.rigid,
            self.type.similarity,
            self.type.affine,
        ):
            return affine_transform(
                data, self.passive[0:-1, 0:-1], offset=self.passive[:-1, -1], **kw
            )
        raise NotImplementedError

    def __mul__(self, other: Transformation):
        if isinstance(other, ScipyHomography):
            if self.passive.shape == other.passive.shape:
                return ScipyHomography(
                    other.passive @ self.passive, warp_options=self._warp_options
                )
            else:
                raise TypeError("Homographies must have same dimensions")
        else:
            raise ValueError("Only concatenation of same types allowed")



from typing import Optional, Mapping
import numpy
from skimage.feature import SIFT, ORB, BRIEF, CENSURE
from skimage.feature import corner_harris, corner_subpix, corner_peaks
from .base import FeatureDetector
from ..features import SciKitImageDescriptorFeatures

__all__ = ["SciKitImageSiftFeatureDetector"]


class SciKitImageSiftFeatureDetector(
    FeatureDetector, registry_id=FeatureDetector.RegistryId("Sift", "SciKitImage")
):
    def __init__(self, feature_options: Optional[Mapping] = None, **kw) -> None:
        if feature_options is None:
            feature_options = dict()
        self._feature_detector = SIFT(**feature_options)
        super().__init__(**kw)

    def find(self, image: numpy.ndarray) -> SciKitImageDescriptorFeatures:
        self._feature_detector.detect_and_extract(image)
        keypoints = self._feature_detector.keypoints
        descriptors = self._feature_detector.descriptors

        if self._mask is not None:
            masked = _mask_keypoints(self._mask, keypoints)
            keypoints = keypoints[masked]
            descriptors = descriptors[masked]

        return SciKitImageDescriptorFeatures(keypoints, descriptors)


class SciKitImageOrbFeatureDetector(
    FeatureDetector, registry_id=FeatureDetector.RegistryId("Orb", "SciKitImage")
):
    def __init__(self, feature_options: Optional[Mapping] = None, **kw) -> None:
        if feature_options is None:
            feature_options = dict()
        self._feature_detector = ORB(**feature_options)
        super().__init__(**kw)

    def find(self, image: numpy.ndarray) -> SciKitImageDescriptorFeatures:
        self._feature_detector.detect_and_extract(image)
        keypoints = self._feature_detector.keypoints
        descriptors = self._feature_detector.descriptors

        if self._mask is not None:
            masked = _mask_keypoints(self._mask, keypoints)
            keypoints = keypoints[masked]
            descriptors = descriptors[masked]

        return SciKitImageDescriptorFeatures(keypoints, descriptors)


class SciKitImageCensureFeatureDetector(
    FeatureDetector,
    registry_id=FeatureDetector.RegistryId("Censure", "SciKitImage"),
    register=False,
):
    def __init__(
        self,
        feature_options: Optional[Mapping] = None,
        descriptor_options: Optional[Mapping] = None,
        **kw
    ) -> None:
        if feature_options is None:
            feature_options = dict()
        if descriptor_options is None:
            descriptor_options = dict()
        self._feature_detector = CENSURE(**feature_options)
        self._descriptor_extractor = BRIEF(**descriptor_options)
        super().__init__(**kw)

    def find(self, image: numpy.ndarray) -> SciKitImageDescriptorFeatures:
        self._feature_detector.detect(image)
        keypoints = self._feature_detector.keypoints
        if self._mask is not None:
            masked = _mask_keypoints(self._mask, keypoints)
            keypoints = keypoints[masked]
        self._descriptor_extractor.extract(image, keypoints)
        return SciKitImageDescriptorFeatures(
            keypoints, self._descriptor_extractor.descriptors
        )


class SciKitImageHarrisFeatureDetector(
    FeatureDetector,
    registry_id=FeatureDetector.RegistryId("Harris", "SciKitImage"),
    register=False,
):
    def __init__(
        self,
        feature_options: Optional[Mapping] = None,
        descriptor_options: Optional[Mapping] = None,
        **kw
    ) -> None:
        if feature_options is None:
            feature_options = dict()
        feature_options.setdefault("min_distance", 5)
        feature_options.setdefault("threshold_rel", 0.01)
        if descriptor_options is None:
            descriptor_options = dict()
        self._feature_options = feature_options
        self._descriptor_extractor = BRIEF(**descriptor_options)
        super().__init__(**kw)

    def find(self, image: numpy.ndarray) -> SciKitImageDescriptorFeatures:
        keypoints = corner_peaks(corner_harris(image), **self._feature_options)
        keypoints = corner_subpix(image, keypoints, window_size=9)
        keypoints = keypoints[~numpy.isnan(keypoints[:, 0])]
        if self._mask is not None:
            masked = _mask_keypoints(self._mask, keypoints)
            keypoints = keypoints[masked]
        self._descriptor_extractor.extract(image, keypoints)
        return SciKitImageDescriptorFeatures(
            keypoints, self._descriptor_extractor.descriptors
        )


def _mask_keypoints(mask: numpy.ndarray, keypoints: numpy.ndarray):
    kp0 = numpy.round(keypoints[:, 0]).astype(numpy.int32)
    kp1 = numpy.round(keypoints[:, 1]).astype(numpy.int32)
    return mask[(kp0, kp1)].astype(bool)



import numpy as np
from typing import Optional
from skimage.filters import (
    window,
    difference_of_gaussians,
    sobel,
    gaussian,
    sato,
    meijering,
    median,
)

from enum import Enum


class WindowType(str, Enum):
    boxcar = "boxcar"
    triang = "triang"
    blackman = "blackman"
    hamming = "hamming"
    hann = "hann"
    bartlett = "bartlett"
    flattop = "flattop"
    parzen = "parzen"
    bohman = "bohman"
    blackmanharris = "blackmanharris"
    nuttall = "nuttall"
    barthann = "barthann"
    cosine = "cosine"
    exponential = "exponential"
    tukey = "tukey"
    taylor = "taylor"
    lanczos = "lanczos"


class FilterType(str, Enum):
    gaussian = "gaussian"
    median = "median"
    sobel = "sobel"
    sato = "sato"
    meijering = "meijering"


def preprocess(
    image: np.ndarray,
    apply_filter: Optional[FilterType] = None,
    filter_parameter: float = 1,
    apply_low_pass: float = 0.0,
    apply_high_pass: float = 0.0,
    apply_window: Optional[WindowType] = None,
    pin_range: bool = False,
):
    if filter_parameter == 0:
        filter_parameter = 1
    if apply_filter == "gaussian":
        image = gaussian(image, filter_parameter)
    if apply_filter == "median":
        image = median(image, np.full((3, 3), 1))
    elif apply_filter == "sobel":
        image = sobel(image)
    elif apply_filter == "sato":
        image = sato(image, np.arange(1, filter_parameter * 2, 2))
    elif apply_filter == "meijering":
        image = meijering(image, np.arange(1, filter_parameter * 2, 2))

    if apply_low_pass > 0 and apply_high_pass > apply_low_pass:
        image = difference_of_gaussians(image, apply_low_pass, apply_high_pass)
    elif apply_low_pass > 0:
        image = difference_of_gaussians(image, apply_low_pass, max(image.shape))
    elif apply_high_pass > 0:
        image = difference_of_gaussians(image, 0, apply_high_pass)

    if pin_range:
        image = (image - image.min()) / (image.max() - image.min())

    if apply_window:
        image = image * window(apply_window, image.shape)

    return image



from typing import Optional, Sequence, Dict
import numpy
from skimage.transform import warp
from skimage.transform import EuclideanTransform
from skimage.transform import SimilarityTransform
from skimage.transform import AffineTransform
from skimage.transform import ProjectiveTransform

from .base import Transformation
from .homography import Homography, reverse_indices
from .lstsq import calc_translation

__all__ = ["SciKitImageHomography"]


class ShiftTransform(ProjectiveTransform):
    def __init__(self, matrix=None, translation=None, dimensionality=2):
        params_given = translation is not None

        if params_given and matrix is not None:
            raise ValueError(
                "You cannot specify the transformation matrix and"
                " the implicit parameters at the same time."
            )
        elif matrix is not None:
            if matrix.shape[0] != matrix.shape[1]:
                raise ValueError("Invalid shape of transformation matrix.")
            self.params = matrix
        elif params_given:
            if translation is None:
                translation = (0,) * dimensionality
            self.params[0:dimensionality, dimensionality] = translation
        else:
            # default to an identity transform
            self.params = numpy.eye(dimensionality + 1)

    def estimate(self, src: numpy.ndarray, dst: numpy.ndarray) -> bool:
        self.params[...] = calc_translation(src.T, dst.T)

    @property
    def translation(self):
        return self.params[0:2, 2]


class SciKitImageHomography(
    Homography, registry_id=Homography.RegistryId("Homography", "SciKitImage")
):
    def __init__(self, *args, warp_options: Optional[Dict] = None, **kw) -> None:
        if warp_options is None:
            warp_options = dict()
        self._warp_options = warp_options

        super().__init__(*args, **kw)

        self._switched_passive = reverse_indices(self.passive)
        if self._type == self._type.identity:
            self._sc_passive = EuclideanTransform(matrix=self._switched_passive)
        elif self._type == self._type.translation:
            self._sc_passive = EuclideanTransform(matrix=self._switched_passive)
        elif self._type == self._type.rigid:
            self._sc_passive = EuclideanTransform(matrix=self._switched_passive)
        elif self._type == self._type.similarity:
            self._sc_passive = SimilarityTransform(matrix=self._switched_passive)
        elif self._type == self._type.affine:
            self._sc_passive = AffineTransform(matrix=self._switched_passive)
        elif self._type == self._type.projective:
            self._sc_passive = ProjectiveTransform(matrix=self._switched_passive)
        else:
            raise ValueError(f"'{self._type}' not supported")
        self._sc_active = self._sc_passive.inverse

    def apply_coordinates(self, coord: Sequence[numpy.ndarray]) -> numpy.ndarray:
        """
        :param coord: shape `(N, M)`
        :returns: shape `(N, M)`
        """
        res = self._sc_active(numpy.transpose(coord[::-1]))
        return numpy.transpose(res)[::-1]

    def apply_data(
        self,
        data: numpy.ndarray,
        offset: Optional[numpy.ndarray] = None,
        shape: Optional[numpy.ndarray] = None,
        cval=numpy.nan,
        interpolation_order: int = 1,
    ) -> numpy.ndarray:
        """
        :param data: shape `(N1, N2, ..., M1, M2, ...)` with `len((N1, N2, ...)) = N`
        :param offset: shape `(N,)`
        :param shape: shape `(N,) = [N1', N2', ...]`
        :param cval: missing value
        :param interpolation_order: order of interpolation: 0 is nearest neighbor, 1 is bilinear,...
        :returns: shape `(N1', N2', ..., M1, M2, ...)`
        """
        kw = dict(self._warp_options)
        if shape is not None:
            kw["output_shape"] = shape
        if offset is not None:
            kw["offset"] = offset
        if cval is not None:
            kw["cval"] = cval
        if interpolation_order in [0, 1, 3]:
            kw["order"] = interpolation_order
        else:
            raise ValueError(
                "This type of interpolation is not supported with ScikitImage"
            )
        # strange behaviour for ints with higher interpolation order
        if numpy.issubdtype(data.dtype, numpy.integer):
            data = data.astype(numpy.float64)
        # TODO: offset
        return warp(data, self._sc_passive, **kw)

    def __mul__(self, other: Transformation):
        if isinstance(other, SciKitImageHomography):
            if self.passive.shape == other.passive.shape:
                return SciKitImageHomography(
                    other.passive @ self.passive, warp_options=self._warp_options
                )
            else:
                raise TypeError("Homographies must have same dimensions")
        else:
            raise ValueError("Only concatenation of same types allowed")



import numpy
from ..features.features import NumpyKeypointFeatures
from ..features.matching.numpy_backend import match_keypoints
from .data import data_for_registration


def test_match_keypoints():
    image = data_for_registration.images("translation", nimages=1)[0][0]
    rs = numpy.random.RandomState(seed=100)
    p0 = rs.uniform(0, image.shape[0], 10)
    p1 = rs.uniform(0, image.shape[1], 10)
    coordinates = numpy.array([p0, p1])
    features1 = NumpyKeypointFeatures(coordinates.T)

    ridx = numpy.arange(10)
    rs.shuffle(ridx)
    features2 = NumpyKeypointFeatures(coordinates.T[ridx])

    idx = match_keypoints(features1, features1, image, image)
    idx1, idx2 = idx.T
    sidx = numpy.argsort(idx1)
    numpy.testing.assert_array_equal(idx1[sidx], numpy.arange(10))
    numpy.testing.assert_array_equal(idx2[sidx], numpy.arange(10))

    idx = match_keypoints(features1, features2, image, image)
    idx1, idx2 = idx.T
    sidx = numpy.argsort(idx1)
    numpy.testing.assert_array_equal(idx1[sidx], numpy.arange(10))
    numpy.testing.assert_array_equal(idx2[sidx], numpy.argsort(ridx))



from typing import Optional, Tuple, Sequence
import numpy

from enum import Enum


class fftConvention(str, Enum):
    numpy = "numpy"
    idl = "idl"


FFT_FREQ_CONVENTION = fftConvention.numpy
# n: number of data points in real space
# d: pixel spacing in real space
#    Convention numpy:
#     frequency = [0,...,k,-k-1,...-1]/(n.d)  (n is even, k = n/2-1  , -k-1 = -n//2, nfreq = k+1 + k+1 = n )
#                 [0,...,k,-k,...,-1]/(n.d)   (n is odd,  k = (n-1)/2, -k   = -n//2, nfreq = k+1 + k = n )
#                 k = (n+1)//2-1
#
#    Convention IDL:
#     frequency = [0,...,k,-k+1,...-1]/(n.d)  (n is even, k = n/2    , -k+1 = -n//2+1, nfreq = k+1 + k-1 = n )
#                 [0,...,k,-k,...,-1]/(n.d)   (n is odd,  k = (n-1)/2, -k   = -n//2  , nfreq = k+1 + k = n )
#                 k = (n+1)//2-(n mod 2)

FFT_NORM_CONVENTION = fftConvention.numpy
# n: number of data points in real space
# Convention numpy:
#   FT  = sum(exp(-2.pi.i.u.x))
#   IFT = sum(exp(2.pi.i.u.x))/n
# Convention IDL ():
#   FT  = sum(exp(-2.pi.i.u.x))/n
#   IFT = sum(exp(2.pi.i.u.x))


def fft_freqind(
    n: int, freqconvention: fftConvention = FFT_FREQ_CONVENTION
) -> Tuple[int, int]:
    """Calculate frequency range of an fft (multiplied by n.d)

    :param n: number of data points in real space
    :param freqconvention: even frequency convention
    """
    # frequencies = [0,...,imax,imin,...,-1]/(n.d)
    if freqconvention == fftConvention.idl:
        imin = -(n // 2) + (1 - (n % 2))
        imax = (n + 1) // 2 - (n % 2)
    else:
        imin = -(n // 2)
        imax = (n + 1) // 2 - 1
    return imin, imax


def fftfreq(
    n: int,
    d: float = 1,
    centered: bool = False,
    freqconvention: fftConvention = FFT_FREQ_CONVENTION,
) -> numpy.ndarray:
    """Fourier space with zero frequency first

    :param n: number of real space data points
    :param d: real space data point spacing
    :param centered: zero frequency in the middle
    :param freqconvention: even frequency convention
    """
    imin, imax = fft_freqind(n, freqconvention=freqconvention)
    if centered:
        freq = numpy.arange(imin, imax + 1, dtype=int)
    else:
        freq = numpy.empty(n, dtype=int)
        numpyos = imax + 1
        freq[:numpyos] = numpy.arange(numpyos, dtype=int)
        freq[numpyos:] = numpy.arange(imin, 0, dtype=int)
    return freq / float(n * d)


def fftshift(
    sigft: numpy.array, freqconvention: fftConvention = FFT_FREQ_CONVENTION
) -> numpy.ndarray:
    """Shift zero frequency to the middle

    :param sigft: signal in Fourier space
    :param freqconvention: even frequency convention
    """
    dim = numpy.array(sigft.shape)
    _, imax = fft_freqind(dim, freqconvention=freqconvention)
    numpyos = imax + 1
    out = sigft.copy()
    for k in range(len(dim)):
        ind = numpy.empty(dim[k], int)
        off = dim[k] - numpyos[k]
        ind[:off] = numpy.arange(numpyos[k], dim[k], dtype=int)
        ind[off:] = numpy.arange(numpyos[k], dtype=int)
        out = numpy.take(out, ind, axis=k)
    return out


def ifftshift(
    sigft: numpy.array, freqconvention: fftConvention = FFT_FREQ_CONVENTION
) -> numpy.ndarray:
    """Shift zero frequency to zero

    :param sigft: signal in Fourier space
    :param freqconvention: even frequency convention
    """
    dim = numpy.array(sigft.shape)
    imin, _ = fft_freqind(dim, freqconvention=freqconvention)
    numpyos = -imin
    out = sigft.copy()
    for k in range(len(dim)):
        ind = numpy.empty(dim[k], int)
        off = dim[k] - numpyos[k]
        ind[:off] = numpy.arange(numpyos[k], dim[k], dtype=int)
        ind[off:] = numpy.arange(numpyos[k], dtype=int)
        out = numpy.take(out, ind, axis=k)
    return out


def _dft1(
    arr,
    dx: float = 1,
    x0: int = 0,
    x1: Optional[int] = None,
    u: Sequence = tuple(),
    centered: bool = False,
    inverse: bool = False,
    normconvention: fftConvention = FFT_NORM_CONVENTION,
    cval=numpy.nan,
) -> numpy.ndarray:
    """Fourier transform with fixed frequencies

    :param arr: array in real or Fourier space
    :param dx: real space data point spacing
    :param x0: real space start index
    :param x1: real space end index
    :param u: Fourier space
    :param centered: zero frequency in the middle
    :param inverse: inverse Fourier transform
    :param normconvention: fft normalization
    :param cval: invalid pixel value
    """
    arr = _cval_to_zero(arr, cval)
    if dx == 1 and x0 == 0 and x1 is None and len(u) == 0 and not centered:
        if inverse:
            ret = numpy.fft.ifft(arr)
            if normconvention == fftConvention.idl:
                ret *= len(arr)
        else:
            ret = numpy.fft.fft(arr)
            if normconvention == fftConvention.idl:
                ret /= len(arr)
    else:
        # Real space
        if x1 is None:
            x1 = len(arr) - 1
        x = numpy.arange(x0, x1 + 1) * dx

        # Fourier space
        if len(u) == 0:
            u = fftfreq(len(arr), dx, centered=centered)

        # Check dimensions
        if inverse:
            if len(u) != len(arr):
                raise ValueError(
                    "Number of frequencies should be equal to the number of data points in Fourier space"
                )
            c = 2j * numpy.pi
            ret = numpy.exp(c * numpy.outer(x, u)).dot(arr)  # nx x nu x nu
        else:
            if len(x) != len(arr):
                raise ValueError(
                    "Number of times should be equal to the number of data points in real space"
                )
            c = -2j * numpy.pi
            ret = numpy.exp(c * numpy.outer(u, x)).dot(arr)  # nu x nx x nx

        # Normalization:
        if (normconvention == fftConvention.idl) ^ inverse:
            ret /= len(u)

    return ret


def _dft2(
    arr: numpy.ndarray,
    dx: float = 1,
    x0: int = 0,
    x1: Optional[float] = None,
    u: Sequence = tuple(),
    dy: float = 1,
    y0: float = 0,
    y1: Optional[float] = None,
    v: Sequence = tuple(),
    centered: bool = False,
    inverse: bool = False,
    normconvention: fftConvention = FFT_NORM_CONVENTION,
    cval=numpy.nan,
) -> numpy.ndarray:
    """Fourier transform with fixed frequencies

    Sub-region inverse Fourier transform with subpixel interpolation using the matrix for of the 2D-DFT
        Manuel Guizar-Sicairos, Samuel T. Thurman, and James R. Fienup,
        "Efficient subpixel image registration algorithms,"
        Optics Letters 33, 156-158 (2008).

    :param arr: array in real or Fourier space
    :param dx: real space data point spacing
    :param x0: real space start index
    :param x1: real space end index
    :param u: Fourier space
    :param dy: real space data point spacing
    :param y0: real space start index
    :param y1: real space end index
    :param v: Fourier space
    :param centered: zero frequency in the middle
    :param inverse: inverse Fourier transform
    :param normconvention: fft normalization
    :param cval: invalid pixel value
    """
    arr = _cval_to_zero(arr, cval)
    if (
        dx == 1
        and x0 == 0
        and x1 is None
        and len(u) == 0
        and dy == 1
        and y0 == 0
        and y1 is None
        and len(v) == 0
        and not centered
    ):
        if inverse:
            ret = numpy.fft.ifft2(arr)
            if normconvention == fftConvention.idl:
                ret *= arr.shape[0] * arr.shape[1]
        else:
            ret = numpy.fft.fft2(arr)
            if normconvention == fftConvention.idl:
                ret /= arr.shape[0] * arr.shape[1]
    else:
        # Real space
        if x1 is None:
            x1 = arr.shape[1] - 1
        if y1 is None:
            y1 = arr.shape[0] - 1
        x = numpy.arange(x0, x1 + 1) * dx
        y = numpy.arange(y0, y1 + 1) * dy

        # Fourier space
        if len(u) == 0:
            u = fftfreq(arr.shape[1], dx, centered=centered)
        if len(v) == 0:
            v = fftfreq(arr.shape[0], dy, centered=centered)

        # DFT (forward or backward)
        if inverse:
            if len(u) != arr.shape[1] or len(v) != arr.shape[0]:
                raise ValueError(
                    "Number of frequencies should be equal to the number of data points in Fourier space"
                )
            c = 2j * numpy.pi
            col_kernel = numpy.exp(c * u[:, None].dot(x[None, :]))  # nu x nx
            row_kernel = numpy.exp(c * y[:, None].dot(v[None, :]))  # ny x nv
            # ny x nv . nv x nu . nu x nx
        else:
            if len(x) != arr.shape[1] or len(y) != arr.shape[0]:
                raise ValueError(
                    "Number of times should be equal to the number of data points in Fourier space"
                )
            c = -2j * numpy.pi
            col_kernel = numpy.exp(c * x[:, None].dot(u[None, :]))  # nx x nu
            row_kernel = numpy.exp(c * v[:, None].dot(y[None, :]))  # nv x ny
            # nv x ny . ny x nx . nx x nu
        ret = row_kernel.dot(arr).dot(col_kernel)

        # Normalization:
        if (normconvention == fftConvention.idl) ^ inverse:
            ret /= len(u) * len(v)

    return ret


def fft1(arr: numpy.ndarray, **kwargs) -> numpy.ndarray:
    """Fourier transform with default frequencies

    :param arr: array in real space
    :returns: array in fourier space
    """
    return _dft1(arr, inverse=False, **kwargs)


def fft2(arr: numpy.ndarray, **kwargs) -> numpy.ndarray:
    """Fourier transform with default frequencies

    :param arr: array in real space
    :returns: array in fourier space
    """
    return _dft2(arr, inverse=False, **kwargs)


def fft(arr: numpy.ndarray, **kwargs) -> numpy.ndarray:
    """Fourier transform with default frequencies

    :param arr: array in real space
    :returns: array in fourier space
    """
    if arr.size in arr.shape:
        if arr.ndim > 1:
            arr = arr.flatten()
        return _dft1(arr, inverse=False, **kwargs)
    else:
        return _dft2(arr, inverse=False, **kwargs)


def ifft1(arr: numpy.ndarray, **kwargs) -> numpy.ndarray:
    """Inverse Fourier transform with default frequencies

    :param arr: array in Fourier space
    :returns: array in fourier space
    """
    return _dft1(arr, inverse=True, **kwargs)


def ifft2(arr: numpy.ndarray, **kwargs) -> numpy.ndarray:
    """Inverse Fourier transform with default frequencies

    :param arr: array in real space
    :returns: array in fourier space
    """
    return _dft2(arr, inverse=True, **kwargs)


def ifft(arr: numpy.ndarray, **kwargs) -> numpy.ndarray:
    """Inverse Fourier transform with default frequencies

    :param arr: array in real space
    :returns: array in fourier space
    """
    if arr.size in arr.shape:
        if arr.ndim > 1:
            arr = arr.flatten()
        return _dft1(arr, inverse=True, **kwargs)
    else:
        return _dft2(arr, inverse=True, **kwargs)


def _cval_to_zero(img: numpy.ndarray, cval):
    if cval != 0:
        if numpy.isnan(cval):
            missing = numpy.isnan(img)
        else:
            missing = img == cval
        bmissing = numpy.any(missing)
    else:
        bmissing = False

    if bmissing:
        img = img.copy()
        img[missing] = 0
    return img


def ifft_interpolate(
    arr: numpy.ndarray,
    ROIoffset: Sequence,
    ROIsize: Sequence,
    sampling: int = 1,
    cval=numpy.nan,
) -> numpy.ndarray:
    """Interpolate array in fourier space

    Sub-region inverse Fourier transform with subpixel interpolation
        Manuel Guizar-Sicairos, Samuel T. Thurman, and James R. Fienup,
        "Efficient subpixel image registration algorithms,"
        Optics Letters 33, 156-158 (2008).

    :param arr: array in fourier space
    :param ROIoffset: sub-pixel grid (super-space)
    :param ROIsize: sub-pixel grid (super-space)
    :returns: inverse FFT on sub-pixel grid
    """
    if arr.size in arr.shape:
        # Super space
        x0 = -ROIoffset
        x1 = x0 + ROIsize - 1

        # Frequencies corresponding to super space
        nu = arr.size
        u = fftfreq(nu, d=sampling)

        return ifft1(arr, x0=x0, x1=x1, u=u, cval=cval)
    else:
        # Super space
        x0 = -ROIoffset[1]
        x1 = x0 + ROIsize[1] - 1
        y0 = -ROIoffset[0]
        y1 = y0 + ROIsize[0] - 1

        # Frequencies corresponding to super space
        nu = arr.shape[1]
        nv = arr.shape[0]
        u = fftfreq(nu, d=sampling)
        v = fftfreq(nv, d=sampling)

        return ifft2(arr, x0=x0, x1=x1, u=u, y0=y0, y1=y1, v=v, cval=cval)



from contextlib import contextmanager
from typing import Dict
from AnyQt import QtWidgets
import numpy as np

from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D, ComplexImageView
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from ewokscore import missing_data
from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.bindings import ow_build_opts
from ewoksorange.gui.parameterform import ParameterForm

from ewoksndreg.registry import RegistryId
from ewoksndreg.tasks import Reg2DIntensities
from ewoksndreg.intensities import registration
from ewoksndreg.intensities.types import (
    SitkMetricType,
    SitkOptimizerType,
    KorniaMetricType,
    KorniaOptimizerType,
)
from ewoksndreg.math.filter import FilterType, WindowType, preprocess
from ewoksndreg.math.fft import fft2, ifft2, fftshift
from ewoksndreg.transformation.types import TransformationType
from ewoksndreg.io.input_stack import input_context, InputStack


__all__ = ["OWReg2DIntensities"]


class OWReg2DIntensities(
    OWEwoksWidgetOneThread, **ow_build_opts, ewokstaskclass=Reg2DIntensities
):
    name = "2D Intensity-Based Registration"
    description = "Calculate transformations of a stack of images based on intensities"
    icon = "icons/registration.png"

    _METHOD_OPTIONS = {
        "Numpy": {},
        "SimpleITK": {
            "metric": {
                "value_for_type": list(SitkMetricType.__members__),
                "serialize": str,
            },
            "optimizer": {
                "value_for_type": list(SitkOptimizerType.__members__),
                "serialize": str,
            },
            "pyramid_levels": {"value_for_type": 0},
            "order": {"value_for_type": 0},
            "sampling": {
                "value_for_type": ["none", "random", "regular"],
                "serialize": str,
            },
            "sampling%": {
                "value_for_type": 0.5,
                "serialize": str,
            },
        },
        "Kornia": {
            "metric": {
                "value_for_type": list(KorniaMetricType.__members__),
                "serialize": str,
            },
            "optimizer": {
                "value_for_type": list(KorniaOptimizerType.__members__),
                "serialize": str,
            },
            "pyramid_levels": {"value_for_type": 0},
        },
        "SciKitImage": {
            "normalized": {"value_for_type": False},
            "upsample_factor": {"value_for_type": 0},
            "sim_normalized": {"value_for_type": False},
            "sim_upsample_factor": {"value_for_type": 0},
        },
    }

    _PREPROC_OPTIONS = {
        "apply_filter": {
            "value_for_type": list(FilterType.__members__),
            "serialize": str,
        },
        "filter_parameter": {"value_for_type": 0.1},
        "apply_low_pass": {"value_for_type": 0.1},
        "apply_high_pass": {"value_for_type": 0.1},
        "pin_range": {"value_for_type": False},
        "apply_window": {
            "value_for_type": list(WindowType.__members__),
            "serialize": str,
        },
    }

    def __init__(self):
        super().__init__()
        self._init_control_area()
        self._init_main_area()

    def _init_control_area(self):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        values = self.get_default_input_values(include_missing=True)
        proper_transformations = list(TransformationType.__members__)
        try:
            proper_transformations.remove("composite")
            proper_transformations.remove("identity")
        except ValueError:
            pass
        options = {
            "imagestack": {"value_for_type": "", "select": "h5dataset"},
            "inputs_are_stacks": {"value_for_type": False},
            "transformation_type": {
                "value_for_type": proper_transformations,
                "serialize": str,
            },
            "reference": {"value_for_type": 0},
            "block_size": {"value_for_type": 1},
        }

        for name, kw in options.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )
        self._default_inputs_form.addParameter(
            "mapper",
            value=values["mapper"],
            value_for_type=registration.IntensityMapping.get_subclass_ids(),
            serialize=str,
            value_change_callback=self._mapper_changed,
        )

        layout = self._get_control_layout()
        self._tabs = QtWidgets.QTabWidget(parent=self.controlArea)
        layout.addWidget(self._tabs)

        # preprocessing tab
        self._prep_widget = QtWidgets.QWidget(parent=self.controlArea)
        layout = QtWidgets.QVBoxLayout()
        self._prep_widget.setLayout(layout)
        self._prepform = ParameterForm(parent=self.controlArea)
        for name, kw in self._PREPROC_OPTIONS.items():
            self._prepform.addParameter(
                name,
                value_change_callback=self._preproc_inputs_changed,
                **kw,
            )
        if values["preprocessing_options"]:
            self._prepform.set_parameter_values(values["preprocessing_options"])
        layout.addWidget(self._prepform)
        self._tabs.addTab(self._prep_widget, "Preprocessing")

        # tabs for each of the backends
        self._pforms: Dict[str, ParameterForm] = {}
        self._tab_widgets: Dict[str, QtWidgets.QWidget] = {}
        for mapper in registration.IntensityMapping.get_subclass_ids():
            options = [
                self._METHOD_OPTIONS[key]
                for key in self._METHOD_OPTIONS.keys()
                if key in str(mapper)
            ][0]

            w = QtWidgets.QWidget(parent=self.controlArea)
            layout = QtWidgets.QVBoxLayout()
            w.setLayout(layout)
            self._pforms[str(mapper)] = ParameterForm(parent=w)
            for name, kw in options.items():
                self._pforms[str(mapper)].addParameter(
                    name,
                    value_change_callback=self._other_inputs_changed,
                    **kw,
                )
            layout.addWidget(self._pforms[str(mapper)])
            self._tabs.addTab(w, mapper.backend)
            self._tab_widgets[str(mapper)] = w
        if values["mapper"] and values["method_options"]:
            self._pforms[values["mapper"]].set_parameter_values(
                values["method_options"]
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_data()

    def _preproc_inputs_changed(self):
        kw = {
            key: item
            for key, item in self._prepform.get_parameter_values().items()
            if not missing_data.is_missing_data(item)
        }
        self.update_default_inputs(preprocessing_options=kw)
        self._add_image(self._vtab.currentIndex(), None)

    def _other_inputs_changed(self):
        current_mapper = self._default_inputs_form.get_parameter_value("mapper")
        if current_mapper:
            method_options = self._pforms[current_mapper].get_parameter_values()
            kw = {
                key: item
                for key, item in method_options.items()
                if not missing_data.is_missing_data(item)
            }
            self.update_default_inputs(method_options=kw)
        self._update_data()

    def _mapper_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        mapper = self.get_task_input_value("mapper", None)
        if mapper is None:
            self._tabs.setCurrentWidget(self._prep_widget)
            self._update_data()
            return
        self._update_transfo_types(RegistryId.factory(mapper))
        self._tabs.setCurrentWidget(self._tab_widgets[mapper])
        method_options = self._pforms[mapper].get_parameter_values()
        kw = {
            key: item
            for key, item in method_options.items()
            if not missing_data.is_missing_data(item)
        }
        self.update_default_inputs(method_options=kw)

    def _update_transfo_types(self, mapper: RegistryId):
        cls = registration.IntensityMapping.get_subclass(mapper)
        current_type = str(self.get_task_input_value("transformation_type"))
        allowed_types = cls.SUPPORTED_TRANSFORMATIONS
        transfo_widg = self._default_inputs_form._get_value_widget(
            "transformation_type"
        )
        transfo_widg.clear()
        transfo_widg.addItem("<missing>", missing_data.MISSING_DATA)
        for _type in allowed_types:
            transfo_widg.addItem(str(_type), str(_type))
        if current_type in allowed_types:
            transfo_widg.setCurrentText(current_type)
            self.update_default_inputs(transformation_type=current_type)

    def handleNewSignals(self) -> None:
        self._update_input_data()
        self._update_data()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_data()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()

        self._vtab = QtWidgets.QTabWidget(parent=self.mainArea)

        layout.addWidget(self._vtab)

        self._main_visual = QtWidgets.QWidget(parent=self.mainArea)
        tab_layout = QtWidgets.QVBoxLayout()
        self._main_visual.setLayout(tab_layout)
        self._plot = Plot2D(parent=self._main_visual)
        self._plot.setDefaultColormap(Colormap("viridis"))
        tab_layout.addWidget(self._plot)
        self._vtab.addTab(self._main_visual, "Images")
        self._plot.getMaskToolsDockWidget().sigMaskChanged.connect(self._mask_changed)

        self._fft_visual = QtWidgets.QWidget(parent=self.mainArea)
        tab_layout = QtWidgets.QVBoxLayout()
        self._fft_visual.setLayout(tab_layout)
        self._fftplot = ComplexImageView.ComplexImageView(parent=self._fft_visual)
        self._fftplot.setColormap(Colormap("viridis", normalization=Colormap.LOGARITHM))
        tab_layout.addWidget(self._fftplot)
        self._vtab.addTab(self._fft_visual, "Fourier")

        self._fftprod_visual = QtWidgets.QWidget(parent=self.mainArea)
        tab_layout = QtWidgets.QVBoxLayout()
        self._fftprod_visual.setLayout(tab_layout)
        self._fftprodplot = ComplexImageView.ComplexImageView(
            parent=self._fftprod_visual
        )
        self._fftprodplot.setColormap(Colormap("viridis"))
        tab_layout.addWidget(self._fftprodplot)
        self._vtab.addTab(self._fftprod_visual, "Cross-correlogram")

        self.current_stack = 0
        self.current_image = 0
        self._vtab.currentChanged[int].connect(self._add_image)

        self._stack_slider = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._stack_slider)
        self._stack_slider.valueChanged[int].connect(self._select_stack)
        self._img_slider = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._img_slider)
        self._img_slider.valueChanged[int].connect(self._select_image)
        self._update_data()

    @contextmanager
    def _input_context(self, images=None):
        if images is not None:
            yield images
            return
        try:
            binit = True
            with input_context(self.get_task_input_value("imagestack")) as images:
                binit = False
                yield images
        except TypeError:
            if binit:
                yield [
                    [],
                ]

    def _update_data(self):
        with self._input_context() as images:
            if isinstance(images, InputStack):
                if images.ndim == 3:
                    self._stack_slider.hide()
                    self._img_slider.setMaximum(max(len(images) - 1, 0))
                else:
                    self._stack_slider.show()
                    self._stack_slider.setMaximum(max(len(images) - 1, 0))
                    self._img_slider.setMaximum(max(len(images[0]) - 1, 0))
            self._select_stack(self._stack_slider.value(), images=images)
            self._select_image(self._img_slider.value(), images=images)

    def _update_input_data(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    def _select_image(self, select, images=None):
        self.current_image = select
        self._add_image(self._vtab.currentIndex(), images)

    def _select_stack(self, select, images=None):
        self.current_stack = select
        self._add_image(self._vtab.currentIndex(), images)

    def _add_image(self, tab, images=None):
        with self._input_context(images=images) as images:
            if not isinstance(images, InputStack):
                return
            preprocessing_options = self.get_task_input_value(
                "preprocessing_options", {}
            )
            if images.ndim == 3:
                image = preprocess(images[self.current_image], **preprocessing_options)
            elif images.ndim == 4:
                image = preprocess(
                    images[self.current_stack, self.current_image],
                    **preprocessing_options,
                )
            if tab == 0:
                self._plot.addImage(image, legend="image")
            elif tab == 1:
                self._fftplot.setData(fft2(image, centered=True), False)
            elif tab == 2:
                reference = self.get_task_input_value("reference", 0)
                if images.ndim == 3:
                    if reference == -1:
                        reference = len(images) // 2
                    ref = preprocess(images[reference], **preprocessing_options)
                elif images.ndim == 4:
                    if reference == -1:
                        reference = images.shape[1] // 2
                    ref = preprocess(
                        images[self.current_stack, reference], **preprocessing_options
                    )
                prod = fft2(image) * fft2(ref).conj()
                pcc = ifft2(prod / np.abs(prod))
                self._fftprodplot.setData(fftshift(pcc), False)

    def _mask_changed(self):
        mask = self._plot.getSelectionMask()
        if mask.ndim == 2:
            if np.any(mask):
                self.update_default_inputs(mask=mask)
            else:
                self.update_default_inputs(mask=missing_data.MISSING_DATA)
            self._default_inputs_changed()



import numpy as np
from typing import List, Union


def range_normalization(image: Union[np.ndarray, List]):
    """
    Scale image intensities to fit in the range [0,1]
    """
    image = np.asarray(image)
    return (image - image.min()) / (image.max() - image.min())


def stack_range_normalization(imagestack: np.ndarray):
    return np.stack([range_normalization(image) for image in imagestack])



from contextlib import contextmanager

from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.bindings import ow_build_opts
from ewoksorange.gui.parameterform import ParameterForm

from ewoksndreg.tasks import FluoAllStacks
from ewoksndreg.io.input_stack import input_context, InputStackNumpy


__all__ = ["OWFluoAllStacks"]


class OWFluoAllStacks(
    OWEwoksWidgetOneThread, **ow_build_opts, ewokstaskclass=FluoAllStacks
):
    name = "ID21 Fluo All Stacks"
    description = (
        "Get a dictionary of all stacks using the provided path to the dataset"
    )
    icon = "icons/load_stacks.svg"

    def __init__(self):
        super().__init__()
        self._init_control_area()
        self._init_main_area()
        self.current_stack = 0
        self.current_image = 0

    def _init_control_area(self):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        values = self.get_default_input_values(include_missing=True)

        options = {
            "path": {
                "value_for_type": "",
                "select": "file",
            },
            "exclude": {
                "value_for_type": "",
                "serialize": str,
            },
            "normalize": {
                "value_for_type": True,
            },
        }
        values["normalize"] = True
        for name, kw in options.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_input_data()

    def handleNewSignals(self) -> None:
        self._update_input_data()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_output_data()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()
        self._plot = Plot2D(parent=self.mainArea)
        self._plot.setDefaultColormap(Colormap("viridis"))
        layout.addWidget(self._plot)
        self._slider = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._slider)
        self._slider.valueChanged[int].connect(self._select_output_stack)
        self._slider2 = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._slider2)
        self._slider2.valueChanged[int].connect(self._select_output_image)
        self._update_output_data()

    def _update_input_data(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    @contextmanager
    def _output_context(self, images=None):
        if images is not None:
            yield images
            return
        try:
            binit = True
            with input_context(self.get_task_output_value("imagestacks")) as images:
                binit = False
                yield images
        except TypeError:
            if binit:
                yield list()

    def _update_output_data(self):
        with self._output_context() as images:
            self._slider.setMaximum(max(len(images) - 1, 0))
            if isinstance(images, InputStackNumpy):
                self._slider2.setMaximum(max(images.shape[1] - 1, 0))
            else:
                self._slider2.setMaximum(max(len(images) - 1, 0))
            self._select_output_image(self._slider2.value(), images=images)
            self._select_output_stack(self._slider.value(), images=images)

    def _select_output_image(self, select, images=None):
        with self._output_context(images=images) as images:
            if images:
                self.current_image = select
                self._plot.addImage(
                    images[self.current_stack, select],
                    legend="image",
                )

    def _select_output_stack(self, select, images=None):
        with self._output_context(images=images) as images:
            if images:
                name = self.get_task_output_value("stack_names")[select]
                self.current_stack = select
                self._plot.addImage(
                    images[select, self.current_image],
                    legend="image",
                    ylabel=" ",
                    xlabel=name,
                )



from contextlib import contextmanager

from silx.gui.plot import Plot2D
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.bindings import ow_build_opts
from ewoksorange.gui.parameterform import ParameterForm

from ewoksndreg.tasks import ID16BFileRead
from ewoksndreg.io.input_stack import input_context


__all__ = ["OWID16BFileRead"]


class OWID16BFileRead(
    OWEwoksWidgetOneThread, **ow_build_opts, ewokstaskclass=ID16BFileRead
):
    name = "ID16B File Read"
    description = "Fetch a stack of edf images using their python script"
    icon = "icons/load_stack.svg"

    def __init__(self):
        super().__init__()
        self._init_control_area()
        self._init_main_area()

    def _init_control_area(self):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        values = self.get_default_input_values(include_missing=True)

        options = {
            "scandir": {"value_for_type": "", "select": "directory"},
            "recsuffix": {
                "value_for_type": "",
            },
            "url": {
                "value_for_type": "",
            },
            "normalize": {
                "value_for_type": False,
            },
        }

        for name, kw in options.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_input_data()

    def handleNewSignals(self) -> None:
        self._update_input_data()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_output_data()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()
        self._plot = Plot2D(parent=self.mainArea)
        layout.addWidget(self._plot)
        self._slider = HorizontalSliderWithBrowser(parent=self.mainArea)
        layout.addWidget(self._slider)
        self._slider.valueChanged[int].connect(self._select_output_image)
        self._update_output_data()

    def _update_input_data(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    @contextmanager
    def _output_context(self, images=None):
        if images is not None:
            yield images
            return
        try:
            binit = True
            with input_context(self.get_task_output_value("imagestack")) as images:
                binit = False
                yield images
        except TypeError:
            if binit:
                yield list()

    def _update_output_data(self):
        with self._output_context() as images:
            self._slider.setMaximum(max(len(images) - 1, 0))
            self._select_output_image(self._slider.value(), images=images)

    def _select_output_image(self, select, images=None):
        with self._output_context(images=images) as images:
            if images:
                self._plot.addImage(images[select], legend="image")



import os
import re
import fabio
from glob import glob
import matplotlib.pyplot as plt


def getfiles(basepath, recsuffix):
    ret = {}
    files = glob(os.path.join(basepath, "*", "*.info"))
    energypattern = re.compile(r"Energy=\s+([0-9\.]+)")
    prefixpattern = re.compile(r"Prefix=\s+([^\s]+_)1_")
    for filename in files:
        energy, prefix = None, None
        with open(filename, "r") as f:
            content = f.readlines()
        for s in content:
            s = s.strip()
            m = energypattern.match(s)
            if m:
                energy = float(m.groups()[0])
            m = prefixpattern.match(s)
            if m:
                prefix = m.groups()[0]
        if energy and prefix:
            recfile = os.path.join(basepath, prefix, prefix + recsuffix)
            if os.path.isfile(recfile):
                ret[energy] = recfile

    energies, files = zip(*sorted(ret.items()))
    return list(energies), list(files)


if __name__ == "__main__":
    scandir = "/tmp_14_days/phasexas/25nm/"
    recsuffix = "6_0p05_rec30000_1.edf"
    energies, files = getfiles(scandir, recsuffix)
    scandir = "/tmp_14_days/phasexas/25nm_cut/"
    recsuffix = "6_0p05_rec30000_1.edf"
    energies[41:48], files[41:48] = getfiles(scandir, recsuffix)
    for filename, energy in zip(files, energies):
        with fabio.open(filename) as f:
            image = f.data
            print(energy, image.shape)
            plt.imshow(image)
            plt.show()



"""Data to test registration"""

from typing import Tuple, List
import numpy

try:
    from skimage import data
    from skimage.transform import warp
    from skimage.transform import SimilarityTransform
    from skimage.transform import AffineTransform, ProjectiveTransform
    from skimage.color import rgb2gray
    from skimage.util import img_as_float
    from skimage.filters import gaussian
except ImportError:
    data = None

try:
    import SimpleITK as sitk
except ImportError:
    sitk = False

from ...math.normalization import stack_range_normalization
from ...transformation.types import TransformationType


def images(
    transfo_type: TransformationType,
    shape: Tuple[int, int] = (200, 220),
    nimages: int = 4,
    plot: float = 0,
    name: str = "astronaut",
) -> Tuple[List[numpy.ndarray], List[numpy.ndarray], List[numpy.ndarray]]:
    if data is None:
        raise ModuleNotFoundError("No module named 'skimage'")
    load_image = getattr(data, name)
    image0 = load_image()
    if image0.ndim > 2:
        image0 = rgb2gray(image0)
    image0 = img_as_float(image0)
    image0 = image0[::-1, :]
    full_shape = numpy.array(image0.shape)
    if all(shape):
        sub_shape = numpy.minimum(numpy.array(shape), full_shape)
    else:
        sub_shape = full_shape
    center = (full_shape / 2).astype(int)
    d = (sub_shape / 2).astype(int)
    idx0 = center - d
    idx1 = center + d
    idx = tuple(slice(i0, i1) for i0, i1 in zip(idx0, idx1))

    if transfo_type == TransformationType.identity:
        tform = SimilarityTransform()
    elif transfo_type == TransformationType.translation:
        tform = SimilarityTransform(translation=[2, 3])
    elif transfo_type == TransformationType.rigid:
        tform = SimilarityTransform(rotation=numpy.radians(4))
    elif transfo_type == TransformationType.similarity:
        tform = SimilarityTransform(scale=1.05)
    elif transfo_type == TransformationType.affine:
        tform = AffineTransform(shear=numpy.radians(4))
    elif transfo_type == TransformationType.projective:
        matrix = numpy.array([[1, 0, 0], [0, 1, 0], [0.001, 0.001, 1]])
        tform = ProjectiveTransform(matrix=matrix)
    elif transfo_type == TransformationType.bspline:
        return get_bspline(image0, nimages, idx)
    elif transfo_type == TransformationType.displacement_field:
        return get_deformations(image0, nimages, idx)

    else:
        raise NotImplementedError(transfo_type)

    tbefore = SimilarityTransform(translation=-center[::-1])
    tafter = SimilarityTransform(translation=center[::-1])

    change_orig1 = SimilarityTransform(translation=idx0[::-1])
    change_orig2 = SimilarityTransform(translation=-idx0[::-1])
    tform0 = tbefore + tform + tafter

    image1 = image0.copy()
    tform1 = tform0
    images = [image1[idx]]
    passive = [numpy.identity(3)]
    active = [numpy.identity(3)]
    if plot:
        import matplotlib.pyplot as plt

        fig = plt.figure()
        plt.imshow(images[-1], origin="lower")
        plt.pause(plot)
    for _ in range(1, nimages):
        image1 = warp(image0, tform1, order=3)
        images.append(image1[idx])
        active.append(indexing_order((change_orig1 + tform1 + change_orig2).params))
        passive.append(
            indexing_order(
                numpy.linalg.inv((change_orig1 + tform1 + change_orig2).params)
            )
        )
        if plot:
            fig.clear()
            plt.imshow(images[-1], origin="lower")
            plt.pause(plot)
        tform1 = tform1 + tform0

    return images, active, passive


def get_deformations(
    image: numpy.ndarray, nimages: int, idx: Tuple[slice]
) -> Tuple[List[numpy.ndarray], List[numpy.ndarray], List[numpy.ndarray]]:
    images = [image[idx]]
    if not sitk:
        raise ModuleNotFoundError("No module named 'SimpleITK'")
    simage = sitk.GetImageFromArray(image)
    shape = image.shape
    deformation = (numpy.random.rand(*shape, 2) - 0.5) * 10
    passive = [numpy.zeros_like(deformation)]
    active = [numpy.zeros_like(deformation)]
    for i in range(nimages):
        active.append(gaussian(deformation * (1.5 * i + 1), i / 4 + 1))
        field = sitk.GetImageFromArray(active[-1], True)
        passive.append(sitk.GetArrayFromImage(sitk.InvertDisplacementField(field)))

        displ = sitk.DisplacementFieldTransform(field)
        result = sitk.Resample(
            simage, simage, displ, sitk.sitkBSpline1, 0.0, simage.GetPixelID()
        )
        images.append(sitk.GetArrayFromImage(result)[idx])
    for i in range(len(passive)):
        passive[i] = passive[i][idx]
        active[i] = active[i][idx]
    return images, active, passive


def get_bspline(
    image: numpy.ndarray, nimages: int, idx: Tuple[slice]
) -> Tuple[List[numpy.ndarray], List[numpy.ndarray], List[numpy.ndarray]]:
    if not sitk:
        raise ModuleNotFoundError("No module named 'SimpleITK'")
    images = [image[idx]]

    simage = sitk.GetImageFromArray(images[0])
    shape = image.shape
    spline_order = 3
    mesh_size = [int(i / 20) for i in shape]
    npoints = [m + spline_order for m in mesh_size]

    displacements = numpy.random.rand(numpy.prod(npoints) * 2) - 0.5
    transform = sitk.BSplineTransformInitializer(simage, mesh_size, 3)
    passive = [numpy.zeros((*shape, 2))]
    active = [[*transform.GetCoefficientImages(), 3]]
    for i in range(nimages):
        scaled = displacements * (i + 1)
        transform.SetParameters(scaled)
        active.append([*transform.GetCoefficientImages(), 3])
        field = sitk.TransformToDisplacementField(
            transform,
            outputPixelType=sitk.sitkVectorFloat64,
            size=simage.GetSize(),
            outputOrigin=simage.GetOrigin(),
            outputSpacing=simage.GetSpacing(),
            outputDirection=simage.GetDirection(),
        )

        passive.append(sitk.GetArrayFromImage(sitk.InvertDisplacementField(field)))
        result = sitk.Resample(
            simage, simage, transform, sitk.sitkBSpline1, 0.0, simage.GetPixelID()
        )
        images.append(sitk.GetArrayFromImage(result))

    return images, active, passive


def noisy_imagestacks(
    transfo_type: TransformationType, nstacks: int = 3, nimages: int = 4
) -> numpy.ndarray:
    stack, _, _ = images(transfo_type, nimages=nimages)
    stack = numpy.stack(stack)
    noise = numpy.random.random(stack.shape)
    stacks = [
        stack_range_normalization(stack + noise * i)
        for i in numpy.linspace(0, 0.5, nstacks)
    ]
    return stacks


def indexing_order(matrix: numpy.ndarray) -> numpy.ndarray:
    matrix = matrix.copy()
    matrix[:2, :2] = matrix[:2, :2].T
    matrix[:2, 2] = matrix[:2, 2][::-1]
    return matrix



from ewoksorange.bindings import ows_to_ewoks
from ewokscore import execute_graph

import pytest
import numpy as np

try:
    import Orange.widgets as Orange
except ImportError:
    Orange = None

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files


@pytest.mark.skipif(Orange is None, reason="python version not supported by orange")
def test_sp_multi_without_qt(tmpdir):
    from ewoksndreg.tests.workflow_tests import workflows

    filename = resource_files(workflows) / "sp_multitransfo.ows"
    assert_multi_without_qt(filename, tmpdir)


@pytest.mark.skipif(Orange is None, reason="python version not supported by orange")
def test_sp_multi_with_qt(ewoks_orange_canvas, tmpdir):
    from ewoksndreg.tests.workflow_tests import workflows

    filename = resource_files(workflows) / "sp_multitransfo.ows"
    assert_multi_with_qt(ewoks_orange_canvas, filename, tmpdir)


def assert_multi_without_qt(filename, tmpdir):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    outputs = execute_graph(
        graph, inputs=[], outputs=[{"all": True}], merge_outputs=False
    )
    expected = outputs["1"]["transformations"]
    for actual in outputs["2"]["transformations"]:
        for ac, ex in zip(actual, expected):
            np.testing.assert_allclose(ac.passive, ex.passive, rtol=0.1, atol=0.05)


def assert_multi_with_qt(ewoks_orange_canvas, filename, tmpdir):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_graph(str(filename), inputs=[])
    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=10)
    outputs = {}
    for i in range(1, 4):
        outputs[str(i)] = next(
            ewoks_orange_canvas.widgets_from_name(str(i))
        ).get_task_output_values()
    expected = outputs["1"]["transformations"]
    for actual in outputs["2"]["transformations"]:
        for ac, ex in zip(actual, expected):
            np.testing.assert_allclose(ac.passive, ex.passive, rtol=0.1, atol=0.05)


@pytest.mark.skipif(Orange is None, reason="python version not supported by orange")
def test_sp_single_without_qt(tmpdir):
    from ewoksndreg.tests.workflow_tests import workflows

    filename = resource_files(workflows) / "sp_singletransfo.ows"
    assert_single_without_qt(filename, tmpdir)


@pytest.mark.skipif(Orange is None, reason="python version not supported by orange")
def test_sp_single_with_qt(ewoks_orange_canvas, tmpdir):
    from ewoksndreg.tests.workflow_tests import workflows

    filename = resource_files(workflows) / "sp_singletransfo.ows"
    assert_single_with_qt(ewoks_orange_canvas, filename, tmpdir)


def assert_single_without_qt(filename, tmpdir):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    outputs = execute_graph(
        graph, inputs=[], outputs=[{"all": True}], merge_outputs=False
    )

    for ac, ex in zip(outputs["1"]["transformations"], outputs["3"]["transformations"]):
        np.testing.assert_allclose(ac.passive, ex.passive, rtol=0.1, atol=0.05)


def assert_single_with_qt(ewoks_orange_canvas, filename, tmpdir):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_graph(str(filename), inputs=[])
    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=10)
    outputs = {}
    for i in range(1, 4):
        outputs[str(i)] = next(
            ewoks_orange_canvas.widgets_from_name(str(i))
        ).get_task_output_values()
    for ac, ex in zip(outputs["1"]["transformations"], outputs["3"]["transformations"]):
        np.testing.assert_allclose(ac.passive, ex.passive, rtol=0.1, atol=0.05)



import numpy as np

from typing import Optional
from scipy.stats import entropy


def normalizedMutualInformationMetric(
    image1: np.ndarray,
    image2: np.ndarray,
    bins: int = 100,
    image1_mask: Optional[np.ndarray] = None,
    image2_mask: Optional[np.ndarray] = None,
) -> float:
    """
    Given two images, the mutual information between the images is calculated based on an histogramm with 'bins' bins
    Masks can be given to exclude some parts of the images from the histogramm calculation
    Alternatively, two stacks of images can be given, where the mutual information per image pair is calculated and summed.
    """
    if image1.ndim == 2 and image2.ndim == 2:
        if image1_mask is not None:
            if image1.shape() != image1_mask.shape:
                raise ValueError("Masks must have same shape as image")
            image1 = image1[image1_mask]

        if image2_mask is not None:
            if image2.shape() != image2_mask.shape:
                raise ValueError("Masks must have same shape as image")
            image2 = image2[image2_mask]

        img1 = image1[np.logical_not(np.isnan(image2))]
        img2 = image2[np.logical_not(np.isnan(image2))]

        joint_bins, edgex, edgey = np.histogram2d(
            img1.flatten(), img2.flatten(), bins=bins
        )

        image1_entropy = entropy(np.sum(joint_bins, axis=0))
        image2_entropy = entropy(np.sum(joint_bins, axis=1))
        joint_entropy = entropy(joint_bins.flatten())

        if joint_entropy != 0:
            return (image1_entropy + image2_entropy) / joint_entropy
        else:
            return 0

    elif image1.ndim == 3 and image2.ndim == 3:
        NME_sum = 0
        for i in range(image1.shape[1]):
            NME_sum += normalizedMutualInformationMetric(
                image1[i],
                image2[i],
                bins=bins,
                image1_mask=image1_mask[i],
                image2_mask=image2_mask[i],
            )
        return NME_sum

    else:
        raise ValueError("ndim of both images must be 2 or 3")


def meanSquaredMetric(image1: np.ndarray, image2: np.ndarray) -> np.float64:
    mean = np.nanmean((image1 - image2) ** 2, dtype=np.float64)
    return mean if not np.isnan(mean) else np.finfo(np.float64).max



from enum import Enum


class TransformationType(str, Enum):
    identity = "identity"
    translation = "translation"
    rigid = "rigid"
    similarity = "similarity"
    affine = "affine"
    projective = "projective"
    bspline = "bspline"
    displacement_field = "displacement_field"
    composite = "composite"



from typing import Optional, Sequence, List, Tuple
from ..io.input_stack import InputStack
from ..transformation.base import Transformation
from .features.base import Features
from .detection.base import FeatureDetector
from .matching.base import FeatureMatching
from .mapping.base import FeatureMapping


def detect_features(
    input_stack: InputStack,
    detector: FeatureDetector,
    include: Optional[Sequence[int]] = None,
) -> List[Features]:
    if include is None:
        include = list(range(len(input_stack)))
    return [detector.find(input_stack[i]) for i in include]


def match_features(
    input_stack: InputStack,
    features: List[Features],
    matcher: FeatureMatching,
    include: Optional[Sequence[int]] = None,
    reference: int = 0,
) -> List[Tuple[Features, Features]]:
    if include is None:
        include = list(range(len(input_stack)))
    ref_features = features[reference]
    matches = list()
    ref_image = input_stack[reference]
    for i in include:
        new_image = input_stack[i]
        if i == reference:
            matches.append((None, None))
        else:
            matches.append(
                matcher.match(ref_features, features[i], ref_image, new_image)
            )
    return matches


def transformations_from_features(
    matches: List[Optional[Tuple[Features, Features]]], mapper: FeatureMapping
) -> List[Transformation]:
    transformations = list()
    for ref_features, new_features in matches:
        if ref_features:
            transformations.append(mapper.calculate(new_features, ref_features))
        else:
            transformations.append(mapper.identity())
    return transformations


def calculate_transformations(
    input_stack: InputStack,
    detector: FeatureDetector,
    matcher: FeatureMatching,
    mapper: FeatureMapping,
    include: Optional[Sequence[int]] = None,
    reference: int = 0,
) -> List[Transformation]:
    features = detect_features(input_stack, detector, include=include)
    matches = match_features(
        input_stack, features, matcher, include=include, reference=reference
    )
    return transformations_from_features(matches, mapper)



from typing import Optional, Sequence, List, Dict
from ..io.input_stack import InputStack
from ..transformation.base import Transformation
from ..math.filter import preprocess
from .base import IntensityMapping


def calculate_transformations(
    input_stack: InputStack,
    mapper: IntensityMapping,
    include: Optional[Sequence[int]] = None,
    reference: int = 0,
    block_size: int = 1,
    preprocessing_options: Optional[Dict] = None,
) -> List[Transformation]:
    """
    Uses the mapper to calculate the transformation between the reference image and all other images in the stack.

    param input_stack: InputStack of all the images
    param mapper: An IntensityMapper to calculate transformations between the images
    param include: Indices of images to include for the registration. Default: all.
    param reference: The index of the reference image, if -1 the reference will be the middle of the stack
    param block_size: Included images get partitioned into blocks.Every image in one block gets aligned to the first element in the block.The first elements of each block get sequentially aligned to the reference

    """
    if include is None:
        include = list(range(len(input_stack)))
    if block_size <= 0:
        block_size = 1
    if reference == -1:
        reference = len(input_stack) // 2
    if reference < -1 or reference >= len(input_stack):
        raise ValueError(
            f"Please use a valid index between 0 and {len(input_stack)} or -1"
        )
    if preprocessing_options is None:
        preprocessing_options = {}
    # ref_index is the index of the reference in include, as include might not contain all indices
    ref_index = include.index(reference)
    ref_image = preprocess(input_stack[reference], **preprocessing_options)
    dim = ref_image.ndim
    transformations = [mapper.identity(dim)]

    # calculate all transformations up to reference
    # counter holds the position in one block, so goes from 0 to block_size-1
    counter = 0
    current_ref_transfo = mapper.identity(dim)
    for i in reversed(include[:ref_index]):
        next_img = preprocess(input_stack[i], **preprocessing_options)
        next_transfo = mapper.calculate(next_img, ref_image) * current_ref_transfo
        counter += 1
        transformations.append(next_transfo)
        if counter % block_size == 0:
            ref_image = next_img
            current_ref_transfo = next_transfo
            counter = 0

    transformations = list(reversed(transformations))

    # calculate all transformations after reference
    ref_image = input_stack[reference]
    counter = 0
    current_ref_transfo = mapper.identity(dim)
    for i in include[ref_index + 1 :]:
        next_img = preprocess(input_stack[i], **preprocessing_options)
        next_transfo = mapper.calculate(next_img, ref_image) * current_ref_transfo
        transformations.append(next_transfo)
        counter += 1
        if counter % block_size == 0:
            ref_image = next_img
            current_ref_transfo = next_transfo
            counter = 0

    return transformations



import numpy as np
import pytest

from ..evaluation.eval_metrics import noisy_eval, peak_eval, smoothness_eval
from .data.data_for_registration import images
from ..math.normalization import stack_range_normalization
from ..transformation.numpy_backend import NumpyHomography
from ..transformation.homography import matrix_from_params


@pytest.mark.parametrize("image", argvalues=["astronaut", "gravel", "cell"])
@pytest.mark.parametrize("noise", ["uniform", "gaussian"])
def test_noisy_metric(image, noise):
    imagestack, active, passive = images("rigid", nimages=5, name=image)
    imagestack = np.stack(imagestack)
    if noise == "uniform":
        noise = np.random.random(imagestack.shape)
    elif noise == "gaussian":
        noise = np.random.normal(size=(imagestack.shape))
    stacks = [imagestack + noise * i * 0.1 for i in range(5)]
    stacks = [stack_range_normalization(stack) for stack in stacks]
    errors = [noisy_eval(stack) for stack in stacks]
    for i in range(len(errors) - 1):
        assert errors[i] < errors[i + 1] + 0.1


@pytest.mark.parametrize("image", argvalues=["astronaut", "gravel", "cell"])
@pytest.mark.parametrize("noise", ["uniform", "gaussian"])
def test_peak_metric(image, noise):
    imagestack, active, passive = images("translation", nimages=5, name=image)
    imagestack = np.stack(imagestack)
    if noise == "uniform":
        noise = np.random.random(imagestack.shape)
    elif noise == "gaussian":
        noise = np.random.normal(size=(imagestack.shape))

    stacks = [imagestack + noise * i * 0.05 for i in range(5)]
    stacks = [stack_range_normalization(stack) for stack in stacks]
    errors = [peak_eval(stack, 0) for stack in stacks]
    for i in range(len(errors) - 1):
        assert errors[i] < errors[i + 1] + 0.01


def test_smoothness_metric():
    img_size = (120, 120)
    matrices = [
        np.array([[1, 0, i], [0, 1, i], [0, 0, 1]], dtype=np.float64) for i in range(10)
    ]
    transformations = [NumpyHomography(matrix) for matrix in matrices]
    bigger_matrices = [
        np.array([[1, 0, 2 * i], [0, 1, 2 * i], [0, 0, 1]], dtype=np.float64)
        for i in range(10)
    ]
    bigger_transformations = [NumpyHomography(matrix) for matrix in bigger_matrices]
    matrices = [
        np.array([[1, 0, i * ((-1) ** i)], [0, 1, i], [0, 0, 1]], dtype=np.float64)
        for i in range(10)
    ]
    jagged_transformations = [NumpyHomography(matrix) for matrix in matrices]
    val1 = smoothness_eval(transformations, img_size)
    val2 = smoothness_eval(bigger_transformations, img_size)
    val3 = smoothness_eval(jagged_transformations, img_size)
    print(val1, val2, val3)
    assert val1 < val2 and val2 < val3


def test_smoothness_metric_rot():
    img_size = (120, 120)
    matrices = [matrix_from_params([i / 20, 0, 0], "rigid") for i in range(10)]
    transformations = [NumpyHomography(matrix) for matrix in matrices]
    bigger_matrices = [matrix_from_params([i / 10, 0, 0], "rigid") for i in range(10)]
    bigger_transformations = [NumpyHomography(matrix) for matrix in bigger_matrices]
    val1 = smoothness_eval(transformations, img_size)
    val2 = smoothness_eval(bigger_transformations, img_size)
    assert val1 < val2


def test_smoothness_metric_size():
    img_size = (50, 50)
    img_size2 = (100, 100)
    matrices = [matrix_from_params([0, i, i], "rigid") for i in range(10)]
    transformations = [NumpyHomography(matrix) for matrix in matrices]
    val1 = smoothness_eval(transformations, img_size)
    val2 = smoothness_eval(transformations, img_size2)
    assert val1 > val2



from ewokscore.task import Task
import numpy as np
from ..tests.data.data_from_h5 import calculate_exclusions, get_stack, get_all_stacks
from ..math.normalization import stack_range_normalization, range_normalization

__all__ = ["FluoStack", "FluoAllStacks"]


class FluoStack(
    Task,
    input_names=["path", "stack"],
    optional_input_names=["exclude", "normalize"],
    output_names=["imagestack", "stack_name", "energies"],
):
    """Fetch a stack of images corresponding to the given counter from the given h5 file"""

    def run(self):
        values = self.get_input_values()
        path = values.pop("path")
        stack = values.pop("stack")
        normalize = values.pop("normalize", False)
        exclude = self.get_input_value("exclude", None)

        exclusions = calculate_exclusions(exclude)

        images, energies = get_stack(path, stack, exclusions)

        images = np.array(images)
        if normalize:
            images = stack_range_normalization(images)
        self.outputs.imagestack = images
        self.outputs.stack_name = stack
        self.outputs.energies = energies


class FluoAllStacks(
    Task,
    input_names=["path"],
    optional_input_names=["exclude", "normalize", "stack_names"],
    output_names=["imagestacks", "stack_names"],
):
    """Fetch all of the different counters from a given h5 file"""

    def run(self):
        values = self.get_input_values()
        path = values.pop("path")
        stack_names = [
            "absorp1",
            "absorp2",
            "absorp3",
            "idet",
            "fluo1",
            "fluo2",
            "fx2_det0_AlKa",
            "fx2_det0_BaL1",
            "fx2_det0_BiM",
            "fx2_det0_CaKa",
            "fx2_det0_CdL",
            "fx2_det0_CeL",
            "fx2_det0_ClKa",
            "fx2_det0_CoKa",
            "fx2_det0_CrKa",
            "fx2_det0_CuKa",
            "fx2_det0_FeKa",
            "fx2_det0_HgM",
            "fx2_det0_ILb",
            "fx2_det0_KKa",
            "fx2_det0_MgKa",
            "fx2_det0_MnKa",
            "fx2_det0_NiKa",
            "fx2_det0_PdL",
            "fx2_det0_PKa",
            "fx2_det0_PtM",
            "fx2_det0_SbL1",
            "fx2_det0_SbL2",
            "fx2_det0_SiKa",
            "fx2_det0_SKa",
            "fx2_det0_SnL",
            "fx2_det0_SnL1",
            "fx2_det0_SrL",
            "fx2_det0_SrL1",
            "fx2_det0_TaM",
            "fx2_det0_TbLa",
            "fx2_det0_TbLb",
            "fx2_det0_TeL2",
            "x2_det0_TeL3",
            "fx2_det0_TiKa",
            "fx2_det0_VKa",
            "fx2_det0_ZnKa",
        ]
        selection = self.get_input_value("stack_names", stack_names)
        if not isinstance(selection[0], str):
            selection = [stack_names[i] for i in selection]
        exclude = values.pop("exclude", None)
        exclusions = calculate_exclusions(exclude)
        images = get_all_stacks(path, selection, exclusions)

        if self.get_input_value("normalize", True):
            for scan in images.keys():
                for counter in images[scan].keys():
                    images[scan][counter] = range_normalization(images[scan][counter])

        self.outputs.stack_names = list(images[list(images)[0]])
        images = [list(scan.values()) for scan in images.values()]
        images = np.array(images).transpose((1, 0, 2, 3))
        self.outputs.imagestacks = images



import numpy as np

from typing import List, Tuple


def crop_stack(stack: np.ndarray) -> np.ndarray:
    """
    Crops the images to exclude all NaN values. Only works for translated images

    param stack: Stack of translated images of form [N,H,W] with nan values in the empty sections

    returns: Stack of cropped images of form [N, H', W']
    """
    boolean = np.isnan(stack)
    full_pixels = boolean.sum(axis=0) == 0
    indices = np.nonzero(full_pixels)
    new_shape = tuple([index.max() - index.min() + 1 for index in indices])
    return stack[:, full_pixels].reshape((-1, *new_shape))


def crop_edges(stack: np.ndarray) -> Tuple[slice, slice]:
    """
    Calculates the slices such that they correspond to cropping out NaN values

    param stack: Stack of translated images of form [N,H,W] with nan values in the empty sections

    returns: two slices which crop the individual images to remove nan values
    """
    boolean = np.isnan(stack)
    full_pixels = boolean.sum(axis=0) == 0
    indices = np.nonzero(full_pixels)
    edges = [slice(index.min(), index.max() + 1) for index in indices]
    return edges


def calc_NaN_edges(image: np.ndarray) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """
    Calculates the biggest subimage that doesn't contain NaN-values
    Usable for situations where loading the entire stack is not feasible
    ----
    param image: Translated image with potential NaN values on the edges

    returns: two tuples containing the 4 borders of the image: ((top,left), (bottom,right))
    """

    boolean = np.isnan(image) == 0
    xindex, yindex = np.nonzero(boolean)
    return (xindex.min(), yindex.min()), (xindex.max() + 1, yindex.max() + 1)


def crop_translations(
    transformations: List,
    include: List[int],
    img_size: Tuple,
    interpolation_order: int = 1,
):
    """
    Calculate the area of the images which doesn't include any NaN values based on the calculated transformations

    param transformations: list of translations
    param include: list of indices of the transformations which should be considered for cropping
    param img_size: image size of the images that will be cropped
    param interpolation_order: The order of the polynomials that will be used to resample the image, as higher orders create more NaN values

    returns: two slices which crop the individual images to remove nan values
    """
    order = interpolation_order
    xshift = [
        int(hom.passive[0, 2]) for i, hom in enumerate(transformations) if i in include
    ]
    yshift = [
        int(hom.passive[1, 2]) for i, hom in enumerate(transformations) if i in include
    ]
    x0, x1 = order, img_size[0] - order
    y0, y1 = order, img_size[1] - order
    if min(xshift) < 0:
        x0 += -min(xshift) + 1
    if max(xshift) > 0:
        x1 -= max(xshift) + 1
    x = slice(x0, x1)

    if min(yshift) < 0:
        y0 += -min(yshift)
    if max(yshift) > 0:
        y1 -= max(yshift)
    y = slice(y0, y1)
    return (x, y)



from typing import Tuple, Optional, Mapping
from silx.opencl.common import ocl
from silx.opencl import sift
from silx.opencl.sift.match import match_py
from .base import FeatureMatching
from ..features import Features
from ..features import SilxDescriptorFeatures


__all__ = ["SilxDescriptorFeatureMatching"]


class SilxDescriptorFeatureMatching(
    FeatureMatching, registry_id=FeatureMatching.RegistryId("Descriptor", "Silx")
):
    def __init__(self, match_options: Optional[Mapping] = None, **kw) -> None:
        if match_options is None:
            match_options = dict()
        self._match_options = match_options
        if ocl is None:
            assert not self._match_options, "No options without opencl"
        self._feature_matcher = None
        super().__init__(**kw)

    def match(
        self, from_features: Features, to_features: Features, *_
    ) -> Tuple[Features, Features]:
        silx_from_features = from_features.as_type(SilxDescriptorFeatures).silx_features
        silx_to_features = to_features.as_type(SilxDescriptorFeatures).silx_features

        nfeatures = max(len(silx_from_features), len(silx_to_features))

        if ocl is None:
            func = match_py
        else:
            if (
                self._feature_matcher is None
                or self._feature_matcher.kpsize < nfeatures
            ):
                self._match_options["size"] = nfeatures
                self._feature_matcher = sift.MatchPlan(**self._match_options)
            func = self._feature_matcher.match
        idx_from, idx_to = func(
            silx_from_features, silx_to_features, raw_results=True
        ).T
        return (from_features[idx_from], to_features[idx_to])



import kornia
import numpy
import torch
import torch.nn.functional as F
import torch.optim as opt

from .base import IntensityMapping
from ..transformation.types import TransformationType
from ..transformation.base import Transformation
from ..transformation.homography import Homography, reverse_indices
from ..transformation.scikitimage_backend import SciKitImageHomography
from .torch_metrics import nmi_loss
from .types import KorniaMetricType, KorniaOptimizerType


class KorniaOptimizationIntensityMapping(
    IntensityMapping,
    registry_id=IntensityMapping.RegistryId("Optimization", "Kornia"),
):
    SUPPORTED_TRANSFORMATIONS = ["translation", "similarity", "projective"]

    def __init__(
        self,
        transfo_type: TransformationType,
        metric: KorniaMetricType = "MeanSquaredError",
        optimizer: KorniaOptimizerType = "RMSprop",
        pyramid_levels: int = 3,
        **kw,
    ) -> None:
        super().__init__(transfo_type, **kw)

        # set transformation type
        if self._type == "projective":
            ktype = "homography"
        else:
            ktype = self._type

        # set metric
        if metric == "MeanAbsoluteError":
            loss_fn = F.l1_loss
        elif metric == "MeanSquaredError":
            loss_fn = F.mse_loss
        elif metric == "MutualInformation":
            loss_fn = nmi_loss
        else:
            raise ValueError(
                f"Invalid Metric, got {metric}, choose one in {[str(item) for item in list(KorniaMetricType)]}"
            )

        # set optimizer
        if optimizer == "SGD":
            optimizer = opt.SGD
        elif optimizer == "Adam":
            optimizer = opt.Adam
        elif optimizer == "RMSprop":
            optimizer = opt.RMSprop
        else:
            raise ValueError(
                f"Invalid Optimizer, got {optimizer}, choose one in {[str(item) for item in list(KorniaOptimizerType)]}"
            )
        self._registrator = kornia.geometry.ImageRegistrator(
            ktype,
            optimizer=optimizer,
            loss_fn=loss_fn,
            pyramid_levels=pyramid_levels,
            tolerance=0.000001,
            num_iterations=200,
        )

    def calculate(
        self, from_image: numpy.ndarray, to_image: numpy.ndarray
    ) -> Homography:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        shp = from_image.shape

        if device == "cuda":
            from_tensor = torch.tensor(from_image, device=device)
            to_tensor = torch.tensor(to_image, device=device)
        else:
            from_tensor = torch.from_numpy(from_image)
            to_tensor = torch.from_numpy(to_image)

        from_tensor = from_tensor.reshape(1, 1, *from_tensor.size())
        to_tensor = to_tensor.reshape(1, 1, *to_tensor.size())

        transformation = self._registrator.register(
            from_tensor.float(), to_tensor.float(), verbose=False
        )

        transformation = reverse_indices(
            numpy.squeeze(transformation.detach().cpu().numpy())
        )
        """
        calculated transformation uses coordinates in range [-1,1]x[-1,1],
        so we need to first transform from our coordinate system to the one used by the transformation (that's pre)
        and afterwards we need to transform back (thats both post transformations)
        """
        pre = numpy.array([[2 / shp[0], 0, -1], [0, 2 / shp[1], -1], [0, 0, 1]])
        post = numpy.array([[1, 0, 1], [0, 1, 1], [0, 0, 1]])
        post2 = numpy.array([[shp[0] / 2, 0, 0], [0, shp[1] / 2, 0], [0, 0, 1]])
        full = post2 @ post @ transformation @ pre
        return SciKitImageHomography(full, transfo_type=self._type)

    def identity(self, dimension: int = 2) -> Transformation:
        return SciKitImageHomography(
            numpy.identity(dimension + 1), TransformationType.identity
        )



import warnings
import numpy

try:
    from scipy.optimize import leastsq
except ImportError:
    leastsq = None


def gaussian(x, y, x0, y0, sx, sy, rho, A):
    num = (
        (x - x0) ** 2 / sx**2
        - 2 * rho / (sx * sy) * (x - x0) * (y - y0)
        + (y - y0) ** 2 / sy**2
    )
    denom = 2 * (1 - rho**2)
    return (
        A / (2 * numpy.pi * sx * sy * numpy.sqrt(1 - rho**2)) * numpy.exp(-num / denom)
    )


def errorf_gaussian(p, x, y, data):
    x0, y0, sx, sy, rho, A = tuple(p)
    return numpy.ravel(gaussian(x, y, x0, y0, sx, sy, rho, A) - data)


def guess_gaussian(x, y, data):
    y0i, x0i = numpy.unravel_index(numpy.argmax(data), data.shape)
    y0 = y[y0i, 0]
    x0 = x[0, x0i]

    xv = x[y0i, :] - x0
    yv = data[y0i, :]
    sx = numpy.sqrt(abs(xv**2 * yv).sum() / yv.sum())
    xv = y[:, x0i] - y0
    yv = data[:, x0i]
    sy = numpy.sqrt(abs(xv**2 * yv).sum() / yv.sum())
    rho = 0.0

    A = data[y0, x0] * 2 * numpy.pi * sx * sy * numpy.sqrt(1 - rho**2)

    return numpy.array([x0, y0, sx, sy, rho, A], dtype=numpy.float32)


def fitgaussian(x, y, data):
    if leastsq is None:
        raise RuntimeError("requires 'scipy'")
    guess = guess_gaussian(x, y, data)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        p, success = leastsq(errorf_gaussian, guess, args=(x, y, data))
        success = success > 0 and success < 5

    return p, success



import numpy
from ...math import fit2d


def test_leastsq():
    nx = 501
    ny = 401
    y, x = numpy.indices((ny, nx))
    x0 = 10
    y0 = ny // 2
    sx = nx // 4
    sy = ny // 4
    rho = 0.5
    A = 1000.0
    p1 = numpy.array([x0, y0, sx, sy, rho, A], dtype=numpy.float32)
    x0, y0, sx, sy, rho, A = tuple(p1)

    data = fit2d.gaussian(x, y, x0, y0, sx, sy, rho, A)

    p2, _ = fit2d.fitgaussian(x, y, data)
    numpy.testing.assert_allclose(p1, p2)



from typing import List
from ewoksorange.bindings import ows_to_ewoks
from ewokscore import execute_graph

import numpy as np
import pytest

try:
    from importlib.resources import files as resource_files
except ImportError:
    from importlib_resources import files as resource_files


pytestmark = pytest.mark.parametrize(
    "reference, block_size",
    [
        (0, 0),
        (29, 1000),
        (59, 5),
        (13, 3),
        pytest.param(0, 1000, marks=pytest.mark.xfail),
    ],
)


def test_strat_workflow_without_qt(tmpdir, reference, block_size):
    from ewoksndreg.tests.workflow_tests import workflows

    filename = resource_files(workflows) / "strategy.ows"
    assert_workflow_without_qt(filename, tmpdir, reference, block_size)


def test_strat_workflow_with_qt(ewoks_orange_canvas, tmpdir, reference, block_size):
    from ewoksndreg.tests.workflow_tests import workflows

    filename = resource_files(workflows) / "strategy.ows"
    assert_workflow_with_qt(
        ewoks_orange_canvas, filename, tmpdir, reference, block_size
    )


def assert_workflow_without_qt(filename, tmpdir, reference, block_size):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    outputs = execute_graph(
        graph,
        inputs=get_inputs(tmpdir, reference, block_size),
        outputs=[{"all": True}],
        merge_outputs=False,
    )
    expected = outputs["0"]["imagestack"][reference]
    for actual in outputs["2"]["imagestack"]:
        expected_flat = expected[np.isfinite(actual)]
        actual_flat = actual[np.isfinite(actual)]
        np.testing.assert_allclose(actual_flat, expected_flat, rtol=0.1, atol=0.05)


def assert_workflow_with_qt(
    ewoks_orange_canvas, filename, tmpdir, reference, block_size
):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_graph(str(filename), inputs=[])
    ewoks_orange_canvas.set_input_values(get_inputs(tmpdir, reference, block_size))
    ewoks_orange_canvas.start_workflow()
    ewoks_orange_canvas.wait_widgets(timeout=10)
    outputs = dict(ewoks_orange_canvas.iter_output_values())
    expected = outputs["0"]["imagestack"][reference]
    for actual in outputs["2"]["imagestack"]:
        expected_flat = expected[np.isfinite(actual)]
        actual_flat = actual[np.isfinite(actual)]
        np.testing.assert_allclose(actual_flat, expected_flat, rtol=0.1, atol=0.05)


def get_inputs(tmpdir, reference, block_size) -> List[dict]:
    return [
        {
            "label": "1",
            "name": "reference",
            "value": reference,
        },
        {
            "label": "1",
            "name": "block_size",
            "value": block_size,
        },
    ]



from contextlib import contextmanager

from silx.gui.plot import Plot2D
from silx.gui.colors import Colormap
from silx.gui.widgets.FrameBrowser import HorizontalSliderWithBrowser

from AnyQt.QtWidgets import QTabWidget, QWidget, QVBoxLayout

from ewoksorange.bindings import OWEwoksWidgetOneThread
from ewoksorange.bindings import ow_build_opts
from ewoksorange.gui.parameterform import ParameterForm

from ewoksndreg.tasks import Reg2DPostEvaluation
from ewoksndreg.io.input_stack import input_context, InputStack


__all__ = ["OWReg2DPostEvaluation"]


class OWReg2DPostEvaluation(
    OWEwoksWidgetOneThread, **ow_build_opts, ewokstaskclass=Reg2DPostEvaluation
):
    name = "Post-Registration Evaluation"
    description = "Choose the best registration based on the calculated transformations for each stack"
    icon = "icons/reg2d_posteval.png"

    def __init__(self):
        super().__init__()
        self._init_control_area()
        self._init_main_area()
        self.current_image = 0
        self.current_stack = 0

    def _init_control_area(self):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        values = self.get_default_input_values(include_missing=True)
        options = {
            "imagestacks": {"value_for_type": "", "select": "h5dataset"},
            "inputs_are_stacks": {"value_for_type": False},
            "chosen_stack": {"value_for_type": -1},
        }

        for name, kw in options.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_input_data()

    def handleNewSignals(self) -> None:
        self._update_input_data()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_output_data()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()
        self._tabs = QTabWidget(parent=self.mainArea)
        layout.addWidget(self._tabs)

        w = QWidget(parent=self.mainArea)
        layout = QVBoxLayout()
        w.setLayout(layout)
        self.cplot = Plot2D(parent=w)
        self._cslider = HorizontalSliderWithBrowser(parent=w)
        layout.addWidget(self.cplot)
        layout.addWidget(self._cslider)
        self._tabs.addTab(w, "Choice")

        w = QWidget(parent=self.mainArea)
        layout = QVBoxLayout()
        w.setLayout(layout)
        self.allplot = Plot2D(parent=w)
        self._stackslider = HorizontalSliderWithBrowser(parent=w)
        self._slider = HorizontalSliderWithBrowser(parent=w)
        layout.addWidget(self.allplot)
        layout.addWidget(self._stackslider)
        layout.addWidget(self._slider)
        self._tabs.addTab(w, "All Options")

        self._stackslider.valueChanged[int].connect(self._select_stack)
        self._slider.valueChanged[int].connect(self._select_image)
        self._cslider.valueChanged[int].connect(self._select_choice_image)

    def _update_input_data(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    @contextmanager
    def _output_context(self, imagestacks=None):
        if imagestacks is not None:
            yield imagestacks
            return
        try:
            binit = True
            with input_context(
                self.get_task_output_value("imagestacks")
            ) as imagestacks:
                binit = False
                yield imagestacks
        except TypeError:
            if binit:
                yield list()

    @contextmanager
    def _choice_context(self, images=None):
        if images is not None:
            yield images
            return
        try:
            binit = True
            with input_context(self.get_task_output_value("imagestack")) as images:
                binit = False
                yield images
        except TypeError:
            if binit:
                yield list()

    def _update_output_data(self):
        with self._output_context() as imagestacks:
            self._stackslider.setMaximum(max(len(imagestacks) - 1, 0))
            if isinstance(imagestacks, InputStack):
                self._slider.setMaximum(max(imagestacks.shape[1] - 1, 0))
            else:
                self._slider.setMaximum(max(len(imagestacks) - 1, 0))

            self._select_stack(self._stackslider.value(), images=imagestacks)
            self._select_image(self._slider.value(), images=imagestacks)

        with self._choice_context() as images:
            self._cslider.setMaximum(max(len(images) - 1, 0))
            self._select_choice_image(self._cslider.value(), images=images)

    def _select_stack(self, select, images=None):
        with self._output_context(imagestacks=images) as images:
            if images:
                self.current_stack = select
                self.allplot.addImage(
                    images[select, self.current_image],
                    legend="image",
                    colormap=Colormap(name="viridis"),
                )

    def _select_image(self, select, images=None):
        with self._output_context(imagestacks=images) as images:
            if images:
                self.current_image = select
                self.allplot.addImage(
                    images[self.current_stack, select],
                    legend="image",
                    colormap=Colormap(name="viridis"),
                )

    def _select_choice_image(self, select, images=None):
        with self._choice_context(images=images) as images:
            if images:
                self.cplot.addImage(
                    images[select],
                    legend="image",
                    colormap=Colormap(name="viridis"),
                )
